import numpy as np
import tensorflow as tf
from dowel import tabular
from garage.tf.misc import tensor_utils

from ast_toolbox.algos import GA
from ast_toolbox.optimizers import DirectionConstraintOptimizer


class GASM(GA):
    """Deep Genetic Algorithm [1]_ with Safe Mutation [2]_.

    Parameters
    ----------
    step_size : float, optional
        The constraint on the KL divergence of each mutation.
    kwargs :
        Keyword arguments passed to `ast_toolbox.algos.ga.GA`.

    References
    ----------
    .. [1] Such, Felipe Petroski, et al. "Deep neuroevolution: Genetic algorithms are a competitive alternative for
    training deep neural networks for reinforcement learning."
     arXiv preprint arXiv:1712.06567 (2017).
    .. [2] Lehman, Joel, et al. "Safe mutations for deep and recurrent neural networks through output gradients."
     Proceedings of the Genetic and Evolutionary Computation Conference. 2018.
    """

    def __init__(
            self,
            step_size=0.01,
            **kwargs):

        self.divergences = np.zeros(kwargs['pop_size'])
        self.optimizer = DirectionConstraintOptimizer()
        super(GASM, self).__init__(**kwargs, step_size=step_size)

    def init_opt(self):
        """Initiate trainer internal tensorflow operations.
        """
        is_recurrent = int(self.policy.recurrent)
        # obs_var = self.env_spec.observation_space.new_tensor_variable(
        #     'obs',
        #     extra_dims=1 + is_recurrent,
        # )
        # action_var = self.env_spec.action_space.new_tensor_variable(
        #     'action',
        #     extra_dims=1 + is_recurrent,
        # )
        if is_recurrent:
            obs_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=[None, None, self.env_spec.observation_space.flat_dim],
                name='obs')
            action_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=[None, None, self.env_spec.action_space.flat_dim],
                name='obs')
        else:
            obs_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=[None, self.env_spec.observation_space.flat_dim],
                name='obs')
            action_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=[None, self.env_spec.action_space.flat_dim],
                name='obs')

        # advantage_var = tensor_utils.new_tensor(
        #     'advantage',
        #     ndim=1 + is_recurrent,
        #     dtype=tf.float32,
        # )

        state_info_vars = {
            k: tf.compat.v1.placeholder(tf.float32, shape=[None] * (1 + is_recurrent) + list(shape), name=k)
            for k, shape in self.policy.state_info_specs
        }
        state_info_vars_list = [state_info_vars[k] for k in self.policy.state_info_keys]

        if is_recurrent:
            valid_var = tf.compat.v1.placeholder(tf.float32, shape=[None, None], name="valid")
        else:
            valid_var = tf.compat.v1.placeholder(tf.float32, shape=[None], name="valid")

        # npath_var = tf.compat.v1.placeholder(tf.int32, shape=(), name="npath")
        npath_var = tf.compat.v1.placeholder(tf.int32, shape=[None], name="npath")  # in order to work with sliced_fn

        actions = self.policy.get_action_sym(obs_var, name='policy_action')
        divergence = tf.reduce_sum(
            tf.reduce_sum(tf.square(actions - action_var), -1) * valid_var) / tf.reduce_sum(valid_var)

        input_list = [
            obs_var,
            action_var,
            # advantage_var,
        ] + state_info_vars_list

        input_list.append(valid_var)
        input_list.append(npath_var)

        self.f_divergence = tensor_utils.compile_function(
            inputs=input_list,
            outputs=divergence,
            log_name="f_divergence",
        )

        self.optimizer.update_opt(
            target=self.policy,
            # leq_constraint=(mean_kl, self.step_size),
            leq_constraint=divergence,  # input max constraint at run time with annealing
            inputs=input_list,
            constraint_name="divergence"
        )
        return dict()

    def extra_recording(self, itr):
        """Record extra training statistics per-iteration.

        Parameters
        ----------
        itr : int
            The iteration number.
        """
        tabular.record('Max Divergence', np.max(self.divergences))
        tabular.record('Min Divergence', np.min(self.divergences))
        tabular.record('Mean Divergence', np.mean(self.divergences))
        return None

    def data2inputs(self, samples_data):
        """Transfer the processed data samples to training inputs

        Parameters
        ----------
        samples_data : dict
            The processed data samples

        Returns
        -------
        all_input_values : tuple
            The input used in training
        """

        all_input_values = (samples_data["observations"], samples_data["actions"])  # ,samples_data["advantages"])
        agent_infos = samples_data["agent_infos"]
        state_info_list = [agent_infos[k] for k in self.policy.state_info_keys]
        all_input_values += tuple(state_info_list)
        # if self.policy.recurrent:
        all_input_values += (samples_data["valids"],)
        npath, max_path_length, _ = all_input_values[0].shape
        if not self.policy.recurrent:
            all_input_values_new = ()
            for (i, item) in enumerate(all_input_values):
                assert item.shape[0] == npath
                assert item.shape[1] == max_path_length
                all_input_values_new += (np.reshape(item, (npath * max_path_length,) + item.shape[2:]),)
            all_input_values_new += (np.ones(npath * max_path_length,) * npath,)
            return all_input_values_new
        else:
            all_input_values += (np.ones(npath) * npath,)
        return all_input_values

    def mutation(self, itr, new_seeds, new_magnitudes, all_paths):
        """Generate new random seeds and magnitudes for the next generation.

        The first self.keep_best seeds are set to no-mutation value (0).

        Parameters
        ----------
        itr : int
            The iteration number.
        new_seeds : :py:class:`numpy.ndarry`
            The original seeds.
        new_magnitudes : :py:class:`numpy.ndarry`
            The original magnitudes.
        all_paths : list[dict]
            The collected paths from the sampler.

        Returns
        -------
        new_seeds : :py:class:`numpy.ndarry`
            The new seeds.
        new_magnitudes : :py:class:`numpy.ndarry`
            The new magnitudes.
        """
        self.seeds = np.copy(new_seeds)
        self.magnitudes = np.copy(new_magnitudes)
        new_seeds[itr + 1, :] = np.random.randint(low=0, high=int(2**16),
                                                  size=(1, self.pop_size))
        for i in range(0, self.keep_best):
            new_seeds[itr + 1, i] = 0
        for p in range(self.pop_size):
            self.set_params(itr, p)
            param_values = self.policy.get_param_values(trainable=True)

            # np.random.seed(int(new_seeds[itr+1,p]))
            # direction = np.random.normal(size=param_values.shape)
            self.np_random.seed(int(new_seeds[itr + 1, p]))
            direction = self.np_random.normal(size=param_values.shape)

            samples_data = all_paths[self.parents[p]]
            all_input_values = self.data2inputs(samples_data)

            new_magnitudes[itr + 1, p], constraint_val = \
                self.optimizer.get_magnitude(
                    direction=direction, inputs=all_input_values, max_constraint_val=self.step_size)
            self.divergences[p] = constraint_val
        return new_seeds, new_magnitudes

    def __getstate__(self):
        """
        Get the internal state.

        Returns
        -------
        data : dict
            The intertal state dict.
        """
        data = self.__dict__.copy()
        del data['f_divergence']
        return data

    def __setstate__(self, state):
        """Set the internal state."""
        self.__dict__ = state
        self._name_scope = tf.name_scope(self.name)
        self.init_opt()

    # for debug
    # def optimize_policy(self, itr, all_paths):
    # 	fitness = self.get_fitness(itr, all_paths)
    # 	self.select_parents(fitness)
    # 	new_seeds = np.zeros_like(self.seeds)
    # 	new_seeds[:,:] = self.seeds[:,self.parents]
    # 	new_magnitudes = np.zeros_like(self.magnitudes)
    # 	new_magnitudes[:,:] = self.magnitudes[:,self.parents]
    # 	if itr+1 < self.n_itr:
    # 		new_seeds, new_magnitudes = self.mutation(itr, new_seeds, new_magnitudes, all_paths)
    # 	self.seeds=new_seeds
    # 	self.magnitudes=new_magnitudes
    # 	# print(self.seeds)
    # 	# print(self.magnitudes)
    # 	for p in range(self.pop_size):
    # 		self.set_params(itr+1,p)
    # 		p_key = self.parents[p]
    # 		all_input_values = self.data2inputs(all_paths[p_key])
    # 		divergence = self.f_divergence(*all_input_values)
    # 		print(divergence)
    # 		self.divergences[p] = divergence
    # 	return dict()



"""
This modules creates a deterministic policy network.

A neural network can be used as policy method in different RL algorithms.
It accepts an observation of the environment and predicts an action.
"""
import torch

from garage.torch.modules import MLPModule
from garage.torch.policies import Policy


class DeterministicMLPPolicy(MLPModule, Policy):
    """
    Implements a deterministic policy network.

    The policy network selects action based on the state of the environment.
    It uses a PyTorch neural network module to fit the function of pi(s).
    """

    def __init__(self, env_spec, **kwargs):
        """
        Initialize class with multiple attributes.

        Args:
            env_spec (garage.envs.env_spec.EnvSpec): Environment specification.
            nn_module (nn.Module): Neural network module in PyTorch.
        """
        self._env_spec = env_spec
        self._obs_dim = env_spec.observation_space.flat_dim
        self._action_dim = env_spec.action_space.flat_dim

        MLPModule.__init__(self,
                           input_dim=self._obs_dim,
                           output_dim=self._action_dim,
                           **kwargs)

    def forward(self, input_val):
        """Forward method."""
        input_val = torch.FloatTensor(input_val)
        return super().forward(input_val)

    def get_action(self, observation):
        """Return a single action."""
        with torch.no_grad():
            x = self.forward(observation.unsqueeze(0))
            return x.squeeze(0).numpy(), dict()

    def get_actions(self, observations):
        """Return multiple actions."""
        with torch.no_grad():
            x = self.forward(observations)
            return x.numpy(), dict()

    def reset(self, dones=None):
        """Reset policy."""
        pass



"""Reshaping Observation for gym.Env."""
import gym
import numpy as np


class ReshapeObservation(gym.Wrapper):
    """
    Reshaping Observation wrapper for gym.Env.

    This wrapper convert the observations into the given shape.

    Args:
        env (gym.Env): The environment to be wrapped.
        shape (list[int]): Target shape to be applied on the observations.
    """

    def __init__(self, env, shape):
        super().__init__(env)
        print(env.observation_space.shape)
        assert np.prod(shape) == np.prod(env.observation_space.shape)
        _low = env.observation_space.low.flatten()[0]
        _high = env.observation_space.high.flatten()[0]
        self._observation_space = gym.spaces.Box(
            _low, _high, shape=shape, dtype=env.observation_space.dtype)
        self._shape = shape

    @property
    def observation_space(self):
        """gym.Env observation space."""
        return self._observation_space

    @observation_space.setter
    def observation_space(self, observation_space):
        self._observation_space = observation_space

    def _observation(self, obs):
        return obs.reshape(self._shape)

    def reset(self):
        """gym.Env reset function."""
        return self._observation(self.env.reset())

    def step(self, action):
        """gym.Env step function."""
        obs, reward, done, info = self.env.step(action)
        return self._observation(obs), reward, done, info



import ast_toolbox.mcts.AdaptiveStressTesting as AST
import ast_toolbox.mcts.AdaptiveStressTestingRandomSeed as AST_RS
from ast_toolbox.algos.mcts import MCTS


class MCTSRS(MCTS):
    """Monte Carlo Tress Search (MCTS) with double progressive widening (DPW) [1]_ using the random seeds as its action space.

    Parameters
    ----------
    seed : int, optional
        The seed used to generate the initial random seed generator.
    rsg_length: int, optional
        The length of the state of the random seed generator. Set it to higher values for extreme large problems.

    References
    ----------
    .. [1] Lee, Ritchie, et al. "Adaptive stress testing of airborne collision avoidance systems."
     2015 IEEE/AIAA 34th Digital Avionics Systems Conference (DASC). IEEE, 2015.
    """

    def __init__(self,
                 seed=0,
                 rsg_length=1,
                 **kwargs):
        self.seed = seed
        self.rsg_length = rsg_length
        super(MCTSRS, self).__init__(**kwargs)

    def init(self):
        """Initiate AST internal parameters
        """
        ast_params = AST.ASTParams(self.max_path_length, self.log_interval, self.log_tabular, self.log_dir, self.n_itr)
        ast_params.rsg_length = self.rsg_length
        ast_params.init_seed = self.seed
        self.ast = AST_RS.AdaptiveStressTestRS(p=ast_params, env=self.env, top_paths=self.top_paths)



"""
This script creates a test that fails when garage.tf.algos.PPO performance is
too low.
"""
import gym
import pytest
import tensorflow as tf

from garage.envs import normalize
from garage.np.baselines import LinearFeatureBaseline
from garage.tf.algos import PPO
from garage.tf.baselines import ContinuousMLPBaseline
from garage.tf.baselines import GaussianMLPBaseline
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import CategoricalMLPPolicy
from garage.tf.policies import GaussianGRUPolicy
from garage.tf.policies import GaussianLSTMPolicy
from garage.tf.policies import GaussianMLPPolicy
from tests.fixtures import snapshot_config, TfGraphTestCase
from tests.fixtures.envs.wrappers import ReshapeObservation


class TestPPO(TfGraphTestCase):

    def setup_method(self):
        super().setup_method()
        self.env = TfEnv(normalize(gym.make('InvertedDoublePendulum-v2')))
        self.policy = GaussianMLPPolicy(
            env_spec=self.env.spec,
            hidden_sizes=(64, 64),
            hidden_nonlinearity=tf.nn.tanh,
            output_nonlinearity=None,
        )
        self.lstm_policy = GaussianLSTMPolicy(env_spec=self.env.spec)
        self.gru_policy = GaussianGRUPolicy(env_spec=self.env.spec)
        self.baseline = GaussianMLPBaseline(
            env_spec=self.env.spec,
            regressor_args=dict(hidden_sizes=(32, 32)),
        )

    # large marker removed to balance test jobs
    # @pytest.mark.large
    def test_ppo_pendulum(self):
        """Test PPO with Pendulum environment."""
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            algo = PPO(env_spec=self.env.spec,
                       policy=self.policy,
                       baseline=self.baseline,
                       max_path_length=100,
                       discount=0.99,
                       lr_clip_range=0.01,
                       optimizer_args=dict(batch_size=32, max_epochs=10))
            runner.setup(algo, self.env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 35

    # large marker removed to balance test jobs
    # @pytest.mark.large
    def test_ppo_with_maximum_entropy(self):
        """Test PPO with maxium entropy method."""
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            algo = PPO(env_spec=self.env.spec,
                       policy=self.policy,
                       baseline=self.baseline,
                       max_path_length=100,
                       discount=0.99,
                       lr_clip_range=0.01,
                       optimizer_args=dict(batch_size=32, max_epochs=10),
                       stop_entropy_gradient=True,
                       entropy_method='max',
                       policy_ent_coeff=0.02,
                       center_adv=False)
            runner.setup(algo, self.env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 35

    # large marker removed to balance test jobs
    # @pytest.mark.large
    def test_ppo_with_neg_log_likeli_entropy_estimation_and_max(self):
        """
        Test PPO with negative log likelihood entropy estimation and max
        entropy method.
        """
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            algo = PPO(env_spec=self.env.spec,
                       policy=self.policy,
                       baseline=self.baseline,
                       max_path_length=100,
                       discount=0.99,
                       lr_clip_range=0.01,
                       optimizer_args=dict(batch_size=32, max_epochs=10),
                       stop_entropy_gradient=True,
                       use_neg_logli_entropy=True,
                       entropy_method='max',
                       policy_ent_coeff=0.02,
                       center_adv=False)
            runner.setup(algo, self.env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 35

    # large marker removed to balance test jobs
    # @pytest.mark.large
    def test_ppo_with_neg_log_likeli_entropy_estimation_and_regularized(self):
        """
        Test PPO with negative log likelihood entropy estimation and
        regularized entropy method.
        """
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            algo = PPO(env_spec=self.env.spec,
                       policy=self.policy,
                       baseline=self.baseline,
                       max_path_length=100,
                       discount=0.99,
                       lr_clip_range=0.01,
                       optimizer_args=dict(batch_size=32, max_epochs=10),
                       stop_entropy_gradient=True,
                       use_neg_logli_entropy=True,
                       entropy_method='regularized',
                       policy_ent_coeff=0.0,
                       center_adv=True)
            runner.setup(algo, self.env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 35

    # large marker removed to balance test jobs
    # @pytest.mark.large
    def test_ppo_with_regularized_entropy(self):
        """Test PPO with regularized entropy method."""
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            algo = PPO(env_spec=self.env.spec,
                       policy=self.policy,
                       baseline=self.baseline,
                       max_path_length=100,
                       discount=0.99,
                       lr_clip_range=0.01,
                       optimizer_args=dict(batch_size=32, max_epochs=10),
                       stop_entropy_gradient=False,
                       entropy_method='regularized',
                       policy_ent_coeff=0.02,
                       center_adv=True,
                       flatten_input=False)
            runner.setup(algo, self.env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 35

    def test_ppo_pendulum_flatten_input(self):
        """Test PPO with CartPole to test observation flattening."""
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            env = TfEnv(
                normalize(ReshapeObservation(gym.make('CartPole-v1'), (2, 2))))
            policy = CategoricalMLPPolicy(
                env_spec=env.spec,
                hidden_nonlinearity=tf.nn.tanh,
            )
            baseline = LinearFeatureBaseline(env_spec=env.spec)
            algo = PPO(env_spec=env.spec,
                       policy=policy,
                       baseline=baseline,
                       max_path_length=100,
                       discount=0.99,
                       gae_lambda=0.95,
                       lr_clip_range=0.2,
                       policy_ent_coeff=0.0,
                       optimizer_args=dict(
                           batch_size=32,
                           max_epochs=10,
                           tf_optimizer_args=dict(learning_rate=1e-3),
                       ))
            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 80

    def teardown_method(self):
        self.env.close()
        super().teardown_method()


class TestPPOContinuousBaseline(TfGraphTestCase):

    @pytest.mark.huge
    def test_ppo_pendulum_continuous_baseline(self):
        """Test PPO with Pendulum environment."""
        with LocalTFRunner(snapshot_config, sess=self.sess) as runner:
            env = TfEnv(normalize(gym.make('InvertedDoublePendulum-v2')))
            policy = GaussianMLPPolicy(
                env_spec=env.spec,
                hidden_sizes=(64, 64),
                hidden_nonlinearity=tf.nn.tanh,
                output_nonlinearity=None,
            )
            baseline = ContinuousMLPBaseline(
                env_spec=env.spec,
                regressor_args=dict(hidden_sizes=(32, 32)),
            )
            algo = PPO(
                env_spec=env.spec,
                policy=policy,
                baseline=baseline,
                max_path_length=100,
                discount=0.99,
                gae_lambda=0.95,
                lr_clip_range=0.2,
                optimizer_args=dict(
                    batch_size=32,
                    max_epochs=10,
                ),
                stop_entropy_gradient=True,
                entropy_method='max',
                policy_ent_coeff=0.02,
                center_adv=False,
            )
            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 100

            env.close()

    @pytest.mark.large
    def test_ppo_pendulum_recurrent_continuous_baseline(self):
        """Test PPO with Pendulum environment and recurrent policy."""
        with LocalTFRunner(snapshot_config) as runner:
            env = TfEnv(normalize(gym.make('InvertedDoublePendulum-v2')))
            policy = GaussianLSTMPolicy(env_spec=env.spec, )
            baseline = ContinuousMLPBaseline(
                env_spec=env.spec,
                regressor_args=dict(hidden_sizes=(32, 32)),
            )
            algo = PPO(
                env_spec=env.spec,
                policy=policy,
                baseline=baseline,
                max_path_length=100,
                discount=0.99,
                gae_lambda=0.95,
                lr_clip_range=0.2,
                optimizer_args=dict(
                    batch_size=32,
                    max_epochs=10,
                ),
                stop_entropy_gradient=True,
                entropy_method='max',
                policy_ent_coeff=0.02,
                center_adv=False,
            )
            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 100

            env.close()


class TestPPOPendulumLSTM(TfGraphTestCase):

    @pytest.mark.large
    def test_ppo_pendulum_lstm(self):
        """Test PPO with Pendulum environment and recurrent policy."""
        with LocalTFRunner(snapshot_config) as runner:
            env = TfEnv(normalize(gym.make('InvertedDoublePendulum-v2')))
            lstm_policy = GaussianLSTMPolicy(env_spec=env.spec)
            baseline = GaussianMLPBaseline(
                env_spec=env.spec,
                regressor_args=dict(hidden_sizes=(32, 32)),
            )
            algo = PPO(
                env_spec=env.spec,
                policy=lstm_policy,
                baseline=baseline,
                max_path_length=100,
                discount=0.99,
                gae_lambda=0.95,
                lr_clip_range=0.2,
                optimizer_args=dict(
                    batch_size=32,
                    max_epochs=10,
                ),
                stop_entropy_gradient=True,
                entropy_method='max',
                policy_ent_coeff=0.02,
                center_adv=False,
            )
            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 80


class TestPPOPendulumGRU(TfGraphTestCase):

    @pytest.mark.large
    def test_ppo_pendulum_gru(self):
        """Test PPO with Pendulum environment and recurrent policy."""
        with LocalTFRunner(snapshot_config) as runner:
            env = TfEnv(normalize(gym.make('InvertedDoublePendulum-v2')))
            gru_policy = GaussianGRUPolicy(env_spec=env.spec)
            baseline = GaussianMLPBaseline(
                env_spec=env.spec,
                regressor_args=dict(hidden_sizes=(32, 32)),
            )
            algo = PPO(
                env_spec=env.spec,
                policy=gru_policy,
                baseline=baseline,
                max_path_length=100,
                discount=0.99,
                gae_lambda=0.95,
                lr_clip_range=0.2,
                optimizer_args=dict(
                    batch_size=32,
                    max_epochs=10,
                ),
                stop_entropy_gradient=True,
                entropy_method='max',
                policy_ent_coeff=0.02,
                center_adv=False,
            )
            runner.setup(algo, env)
            last_avg_ret = runner.train(n_epochs=10, batch_size=2048)
            assert last_avg_ret > 80



import tensorflow as tf

from garage.tf.envs import TfEnv
from garage.tf.policies import CategoricalGRUPolicy
from garage.tf.policies import CategoricalLSTMPolicy
from garage.tf.policies import CategoricalMLPPolicy
from garage.tf.policies import ContinuousMLPPolicy
from garage.tf.policies import GaussianGRUPolicy
from garage.tf.policies import GaussianLSTMPolicy
from garage.tf.policies import GaussianMLPPolicy
from tests.fixtures import TfGraphTestCase
from tests.fixtures.envs.dummy import DummyBoxEnv, DummyDiscreteEnv


class TestDiscretePolicies(TfGraphTestCase):

    def setup_method(self):
        super().setup_method()
        self.env = TfEnv(DummyDiscreteEnv())

    def teardown_method(self):
        self.env.close()
        super().teardown_method()

    def test_categorial_gru_policy(self):
        categorical_gru_policy = CategoricalGRUPolicy(env_spec=self.env,
                                                      hidden_dim=1)
        self.sess.run(tf.compat.v1.global_variables_initializer())

        categorical_gru_policy.reset()

        obs = self.env.observation_space.high
        assert categorical_gru_policy.get_action(obs)

    def test_categorical_lstm_policy(self):
        categorical_lstm_policy = CategoricalLSTMPolicy(env_spec=self.env,
                                                        hidden_dim=1)
        self.sess.run(tf.compat.v1.global_variables_initializer())

        categorical_lstm_policy.reset()

        obs = self.env.observation_space.high
        assert categorical_lstm_policy.get_action(obs)

    def test_categorial_mlp_policy(self):
        categorical_mlp_policy = CategoricalMLPPolicy(env_spec=self.env,
                                                      hidden_sizes=(1, ))
        self.sess.run(tf.compat.v1.global_variables_initializer())

        obs = self.env.observation_space.high
        assert categorical_mlp_policy.get_action(obs)


class TestContinuousPolicies(TfGraphTestCase):

    def setup_method(self):
        super().setup_method()
        self.env = TfEnv(DummyBoxEnv())

    def teardown_method(self):
        self.env.close()
        super().teardown_method()

    def test_continuous_mlp_policy(self):
        continuous_mlp_policy = ContinuousMLPPolicy(env_spec=self.env,
                                                    hidden_sizes=(1, ))
        self.sess.run(tf.compat.v1.global_variables_initializer())

        obs = self.env.observation_space.high
        assert continuous_mlp_policy.get_action(obs)

    def test_gaussian_gru_policy(self):
        gaussian_gru_policy = GaussianGRUPolicy(env_spec=self.env,
                                                hidden_dim=1)
        self.sess.run(tf.compat.v1.global_variables_initializer())

        gaussian_gru_policy.reset()

        obs = self.env.observation_space.high
        assert gaussian_gru_policy.get_action(obs)

    def test_gaussian_lstm_policy(self):
        gaussian_lstm_policy = GaussianLSTMPolicy(env_spec=self.env,
                                                  hidden_dim=1)
        self.sess.run(tf.compat.v1.global_variables_initializer())

        gaussian_lstm_policy.reset()

        obs = self.env.observation_space.high
        assert gaussian_lstm_policy.get_action(obs)

    def test_gaussian_mlp_policy(self):
        gaussian_mlp_policy = GaussianMLPPolicy(env_spec=self.env,
                                                hidden_sizes=(1, ))
        self.sess.run(tf.compat.v1.global_variables_initializer())

        obs = self.env.observation_space.high
        assert gaussian_mlp_policy.get_action(obs)



# Define the class, inherit from the base
class ASTReward(object):
    """Function to calculate the rewards for timesteps when optimizing AST solver policies.

    """

    def __init__(self):
        pass

    def give_reward(self, action, **kwargs):
        """Returns the reward for a given time step.

        Parameters
        ----------
        action : array_like
            Action taken by the AST solver.
        kwargs :
            Accepts relevant info for computing the reward.
        Returns
        -------
        reward : float
            Reward based on the previous action.
        """

        raise NotImplementedError



"""Baseline estimators for TensorFlow-based algorithms."""
from garage.tf.baselines.continuous_mlp_baseline import ContinuousMLPBaseline
from garage.tf.baselines.gaussian_cnn_baseline import GaussianCNNBaseline
from garage.tf.baselines.gaussian_mlp_baseline import GaussianMLPBaseline

__all__ = [
    'ContinuousMLPBaseline',
    'GaussianCNNBaseline',
    'GaussianMLPBaseline',
]



"""Samplers for solving AST formualted RL problems."""
from .ast_vectorized_sampler import ASTVectorizedSampler  # noqa
from .batch_sampler import BatchSampler  # noqa

__all__ = ['ASTVectorizedSampler', 'BatchSampler']



""" Gym environment to turn general AST tasks into garage compatible problems."""
import gym
import numpy as np
from cached_property import cached_property
from garage.envs.base import Step
from garage.envs.env_spec import EnvSpec

from ast_toolbox.rewards import ExampleAVReward
from ast_toolbox.simulators import ExampleAVSimulator


class ASTEnv(gym.Env):
    r""" Gym environment to turn general AST tasks into garage compatible problems.

    Parameters
    ----------
    open_loop : bool
        True if the simulation is open-loop, meaning that AST must generate all actions ahead of time, instead
        of being able to output an action in sync with the simulator, getting an observation back before
        the next action is generated. False to get interactive control, which requires that `blackbox_sim_state`
        is also False.
    blackbox_sim_state : bool
        True if the true simulation state can not be observed, in which case actions and the initial conditions are
        used as the observation. False if the simulation state can be observed, in which case it will be used.
    fixed_init_state : bool
        True if the initial state is fixed, False to sample the initial state for each rollout from the observaation
        space.
    s_0 : array_like
        The initial state for the simulation (ignored if `fixed_init_state` is False)
    simulator : :py:class:`ast_toolbox.simulators.ASTSimulator`
        The simulator wrapper, inheriting from `ast_toolbox.simulators.ASTSimulator`.
    reward_function : :py:class:`ast_toolbox.rewards.ASTReward`
        The reward function, inheriting from `ast_toolbox.rewards.ASTReward`.
    spaces : :py:class:`ast_toolbox.spaces.ASTSpaces`
        The observation and action space definitions, inheriting from `ast_toolbox.spaces.ASTSpaces`.
    """

    def __init__(self,
                 open_loop=True,
                 blackbox_sim_state=True,
                 fixed_init_state=False,
                 s_0=None,
                 simulator=None,
                 reward_function=None,
                 spaces=None):

        # Constant hyper-params -- set by user
        self.open_loop = open_loop
        self.blackbox_sim_state = blackbox_sim_state  # is this redundant?
        self.spaces = spaces
        # These are set by reset, not the user
        self._done = False
        self._reward = 0.0
        self._step = 0
        self._action = None
        self._actions = []
        self._first_step = True
        self.reward_range = (-float('inf'), float('inf'))
        self.metadata = None
        self._cum_reward = 0.0

        if s_0 is None:
            self._init_state = self.observation_space.sample()
        else:
            self._init_state = s_0
        self._fixed_init_state = fixed_init_state
        self.simulator = simulator
        if self.simulator is None:
            self.simulator = ExampleAVSimulator()
        self.reward_function = reward_function
        if self.reward_function is None:
            self.reward_function = ExampleAVReward()

        if hasattr(self.simulator, "vec_env_executor") and callable(getattr(self.simulator, "vec_env_executor")):
            self.vectorized = True
        else:
            self.vectorized = False
        # super().__init__(self)

    def step(self, action):
        r"""
        Run one timestep of the environment's dynamics. When end of episode
        is reached, reset() should be called to reset the environment's internal state.

        Parameters
        ----------
        action : array_like
            An action provided by the environment.

        Returns
        -------
        : :py:func:`garage.envs.base.Step`
            A step in the rollout.
            Contains the following information:
                - observation (array_like): Agent's observation of the current environment.
                - reward (float): Amount of reward due to the previous action.
                - done (bool): Is the current step a terminal or goal state, ending the rollout.
                - actions (array_like): The action taken at the current.
                - state (array_like): The cloned simulation state at the current cell, used for resetting if chosen to start a rollout.
                - is_terminal (bool): Whether or not the current cell is a terminal state.
                - is_goal (bool): Whether or not the current cell is a goal state.
        """
        self._env_state_before_action = self._env_state.copy()

        self._action = action
        self._actions.append(action)
        action_return = self._action

        # Update simulation step
        obs = self.simulator.step(self._action)
        if (obs is None) or (self.open_loop is True) or (self.blackbox_sim_state):
            obs = np.array(self._init_state)

        if self.simulator.is_terminal() or self.simulator.is_goal():
            self._done = True

        # Calculate the reward for this step
        self._reward = self.reward_function.give_reward(
            action=self._action,
            info=self.simulator.get_reward_info())
        self._cum_reward += self._reward

        # Update instance attributes
        self._step = self._step + 1

        self._simulator_state = self.simulator.clone_state()
        self._env_state = np.concatenate((self._simulator_state,
                                          np.array([self._cum_reward]),
                                          np.array([self._step])),
                                         axis=0)

        return Step(observation=obs,
                    reward=self._reward,
                    done=self._done,
                    actions=action_return,
                    state=self._env_state_before_action,
                    is_terminal=self.simulator.is_terminal(),
                    is_goal=self.simulator.is_goal())

    def simulate(self, actions):
        r"""Run a full simulation rollout.

        Parameters
        ----------
        actions : list[array_like]
            A list of array_likes, where each member is the action taken at that step.

        Returns
        -------
        int
            The step of the trajectory where a collision was found, or -1 if a collision was not found.
        dict
            A dictionary of simulation information for logging and diagnostics.
        """
        if not self._fixed_init_state:
            self._init_state = self.observation_space.sample()
        return self.simulator.simulate(actions, self._init_state)

    def reset(self):
        r"""Resets the state of the environment, returning an initial observation.

        Returns
        -------
        observation : array_like
            The initial observation of the space. (Initial reward is assumed to be 0.)
        """
        self._actions = []
        if not self._fixed_init_state:
            self._init_state = self.observation_space.sample()
        self._done = False
        self._reward = 0.0
        self._cum_reward = 0.0
        self._action = None
        self._actions = []
        self._first_step = True
        self._step = 0

        obs = np.array(self.simulator.reset(self._init_state))
        if not self._fixed_init_state:
            obs = np.concatenate((obs, np.array(self._init_state)), axis=0)

        self._simulator_state = self.simulator.clone_state()
        self._env_state = np.concatenate((self._simulator_state,
                                          np.array([self._cum_reward]),
                                          np.array([self._step])),
                                         axis=0)

        return obs

    @property
    def action_space(self):
        r"""Convenient access to the environment's action space.

        Returns
        -------
        : `gym.spaces.Space <https://gym.openai.com/docs/#spaces>`_
            The action space of the reinforcement learning problem.
        """
        if self.spaces is None:
            # return self._to_garage_space(self.simulator.action_space)
            return self.simulator.action_space
        else:
            return self.spaces.action_space

    @property
    def observation_space(self):
        r"""Convenient access to the environment's observation space.

        Returns
        -------
        : `gym.spaces.Space <https://gym.openai.com/docs/#spaces>`_
            The observation space of the reinforcement learning problem.
        """
        if self.spaces is None:
            # return self._to_garage_space(self.simulator.observation_space)
            return self.simulator.observation_space
        else:
            return self.spaces.observation_space

    def log(self):
        r"""Calls the simulator's `log` function.

        """
        self.simulator.log()

    def render(self, **kwargs):
        r"""Calls the simulator's `render` function, if it exists.

        Parameters
        ----------
        kwargs :
            Keyword arguments used in the simulators `render` function.

        Returns
        -------
        None or object
            Returns the output of the simulator's `render` function, or None if the simulator has no `render` function.
        """
        if hasattr(self.simulator, "render") and callable(getattr(self.simulator, "render")):
            return self.simulator.render(**kwargs)
        else:
            return None

    def close(self):
        r"""Calls the simulator's `close` function, if it exists.

        Returns
        -------
        None or object
            Returns the output of the simulator's `close` function, or None if the simulator has no `close` function.
        """
        if hasattr(self.simulator, "close") and callable(getattr(self.simulator, "close")):
            return self.simulator.close()
        else:
            return None

    @cached_property
    def spec(self):
        r"""Returns a garage environment specification.

        Returns
        -------
        :py:class:`garage.envs.env_spec.EnvSpec`
            A garage environment specification.
        """
        return EnvSpec(
            observation_space=self.observation_space,
            action_space=self.action_space)



import csv
import os.path

import matplotlib
import numpy as np
from matplotlib import pyplot as plt
from matplotlib2tikz import save as tikz_save

matplotlib.rcParams.update({'font.size': 10, 'font.family': 'Times New Roman',
                            'text.usetex': True})

n_trial = 10
top_k = 1
batch_size = 1000
max_step = np.inf
max_reward = np.inf
min_reward = -np.inf

exp_name = 'CartpoleNd'
exp_param = 'D1SUT5'
extra_name = ''  # 'hyper'
prepath = "../" + exp_name + "/Data/AST/Lexington/" + exp_param
plot_path = "../" + exp_name + "/Data/Plot/avgtop" + str(top_k) + "/"
# plot_path = "/Users/xiaobaima/Dropbox/SISL/ASTToolbox/IROS/plots/"

policy_groups = [
                [
                    "RandomSearchB2000",
                ],
    [
                    # "TRPOB2000Step0.1","TRPOB2000Step1.0",
                    "TRPOB2000Step5.0",\
                    # "TRPOB50000Step0.1","TRPOB50000Step1.0","TRPOB50000Step5.0",\
                ],\
    [
                    # "TRPOLSTMB2000Step0.1",
                    "TRPOLSTMB2000Step1.0",
                    # "TRPOLSTMB2000Step5.0",\
                    # "TRPOLSTMB50000Step0.1","TRPOLSTMB50000Step1.0","TRPOLSTMB50000Step5.0",\
                ],\
    [
                    # "MCTSRSEc10.0K0.3A0.3",
                    "MCTSRSEc10.0K0.5A0.5",
                    # "MCTSRSEc10.0K0.8A0.8",\
                    # "MCTSASEc10.0K0.3A0.3","MCTSASEc10.0K0.5A0.5","MCTSASEc10.0K0.8A0.8",\
                    # "MCTSBVEc10.0K0.3A0.3","MCTSBVEc10.0K0.5A0.5","MCTSBVEc10.0K0.8A0.8",\
                ]\
]

algos = ["Random Search", "TRPO, Closed-Loop", "TRPO, Open-Loop", "MCTS, Open-Loop"]
parameters = [[""],
              ["step size 0.1"],
              ['$k=\\alpha=0.5,ec=10.0$', '$k=\\alpha=0.5,ec=10.0$', '$k=\\alpha=0.5,ec=10.0$'],
              ]

exp_name_f = 'Cartpole'

exp_param_f = ''

colors = ["black", "blue", "green", "red"]
plot_name = exp_name + '_' + exp_param + 'avgtop' + str(top_k) + 'trial' + str(n_trial) + extra_name


plts = []
legends = []
fig = plt.figure(figsize=(10, 10))

for (group_index, policies) in enumerate(policy_groups):
    best_policy_index = -1
    best_rewards = []
    best_steps = []
    best_last_reward = -np.inf
    for (policy_index, policy) in enumerate(policies):
        print(policy)
        Rewards = []
        min_array_length = np.inf
        for trial in range(n_trial):
            file_path = prepath + '/' + policy + '/' + str(trial) + '/process.csv'
            if os.path.exists(file_path):
                print(trial)
                steps = []
                rewards = []
                with open(file_path) as csv_file:
                    if '\0' in open(file_path).read():
                        print("you have null bytes in your input file")
                        csv_reader = csv.reader(x.replace('\0', '') for x in csv_file)
                    else:
                        csv_reader = csv.reader(csv_file, delimiter=',')

                    for (i, row) in enumerate(csv_reader):
                        if i == 0:
                            entry_dict = {}
                            for index in range(len(row)):
                                entry_dict[row[index]] = index
                        else:
                            if int(row[entry_dict["StepNum"]]) > max_step:
                                break
                            if int(row[entry_dict["StepNum"]]) % batch_size == 0:
                                steps.append(int(row[entry_dict["StepNum"]]))
                                avg_top = 0.0
                                for k in range(top_k):
                                    avg_top += np.clip(float(row[entry_dict["reward " + str(k)]]), min_reward, max_reward)
                                avg_top /= top_k
                                rewards.append(avg_top)
                if len(rewards) < min_array_length:
                    min_array_length = len(rewards)
                Rewards.append(rewards)
        steps = steps[:min_array_length]
        Rewards = [rewards[:min_array_length] for rewards in Rewards]
        if np.mean(Rewards, 0)[-1] > best_last_reward:
            best_last_reward = np.mean(Rewards, 0)[-1]
            best_policy_index = policy_index
            best_rewards = Rewards
            best_steps = steps
    Rewards = best_rewards
    steps = best_steps
    # y = np.mean(Rewards,0)
    # yerr = np.std(Rewards,0)/np.sqrt(n_trial)
    plot, = plt.plot(steps, np.mean(Rewards, 0), color=colors[group_index])
    print(np.max(Rewards))
    # plt.fill_between(steps,y+yerr,y-yerr,linewidth=0,facecolor=colors[group_index],alpha=0.3)
    plts.append(plot)
    # legends.append(algos[group_index]+' '+parameters[group_index][best_policy_index])
    legends.append(algos[group_index])

axes = plt.gca()
# start, end = axes.get_xlim()
# axes.set_xlim([0,2e6])
# axes.xaxis.set_ticks(np.arange(0, 2e6, int(2e6/4)))
# axes.set_yscale('log')
# axes.set_xlim([xmin,xmax])
# axes.set_ylim([-0.15,1.0])
# plt.title(exp_name_f+' '+exp_param_f)
plt.legend(plts, legends)
plt.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
# plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
plt.xlabel('Step Number')
# plt.ylabel('Average Top '+str(top_k) +' Reward')
plt.ylabel('Average Best Return')
fig.savefig(plot_path + plot_name + '.pdf', bbox_inches='tight')
tikz_save(plot_path + plot_name + '.tex')
plt.close(fig)



import pickle
from unittest import mock

import numpy as np
import pytest
import tensorflow as tf

from garage.tf.models import GaussianGRUModel
from tests.fixtures import TfGraphTestCase
from tests.helpers import recurrent_step_gru


class TestGaussianGRUModel(TfGraphTestCase):

    def setup_method(self):
        super().setup_method()
        self.batch_size = 1
        self.time_step = 2
        self.feature_shape = 2
        self.default_initializer = tf.constant_initializer(0.1)

        self.obs_inputs = np.full(
            (self.batch_size, self.time_step, self.feature_shape), 1.)
        self.obs_input = np.full((self.batch_size, self.feature_shape), 1.)

        self.input_var = tf.compat.v1.placeholder(tf.float32,
                                                  shape=(None, None,
                                                         self.feature_shape),
                                                  name='input')
        self.step_input_var = tf.compat.v1.placeholder(
            tf.float32, shape=(None, self.feature_shape), name='step_input')

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    @mock.patch('tensorflow.random.normal')
    def test_std_share_network_output_values(self, mock_normal, output_dim,
                                             hidden_dim):
        mock_normal.return_value = 0.5
        model = GaussianGRUModel(output_dim=output_dim,
                                 hidden_dim=hidden_dim,
                                 std_share_network=True,
                                 hidden_nonlinearity=None,
                                 recurrent_nonlinearity=None,
                                 hidden_w_init=self.default_initializer,
                                 recurrent_w_init=self.default_initializer,
                                 output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(shape=(self.batch_size,
                                                          hidden_dim),
                                                   name='step_hidden',
                                                   dtype=tf.float32)
        (mean_var, step_mean_var, log_std_var, step_log_std_var, step_hidden,
         hidden_init_var, dist) = model.build(self.input_var,
                                              self.step_input_var,
                                              step_hidden_var)

        hidden1 = hidden2 = np.full((self.batch_size, hidden_dim),
                                    hidden_init_var.eval())

        mean, log_std = self.sess.run(
            [mean_var, log_std_var],
            feed_dict={self.input_var: self.obs_inputs})

        for i in range(self.time_step):
            mean1, log_std1, hidden1 = self.sess.run(
                [step_mean_var, step_log_std_var, step_hidden],
                feed_dict={
                    self.step_input_var: self.obs_input,
                    step_hidden_var: hidden1
                })

            hidden2 = recurrent_step_gru(input_val=self.obs_input,
                                         num_units=hidden_dim,
                                         step_hidden=hidden2,
                                         w_x_init=0.1,
                                         w_h_init=0.1,
                                         b_init=0.,
                                         nonlinearity=None,
                                         gate_nonlinearity=None)

            output_nonlinearity = np.full(
                (np.prod(hidden2.shape[1:]), output_dim), 0.1)
            output2 = np.matmul(hidden2, output_nonlinearity)
            assert np.allclose(mean1, output2)
            assert np.allclose(log_std1, output2)
            assert np.allclose(hidden1, hidden2)

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    def test_std_share_network_shapes(self, output_dim, hidden_dim):
        model = GaussianGRUModel(output_dim=output_dim,
                                 hidden_dim=hidden_dim,
                                 std_share_network=True,
                                 hidden_nonlinearity=None,
                                 recurrent_nonlinearity=None,
                                 hidden_w_init=self.default_initializer,
                                 recurrent_w_init=self.default_initializer,
                                 output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(shape=(self.batch_size,
                                                          hidden_dim),
                                                   name='step_hidden',
                                                   dtype=tf.float32)
        (mean_var, step_mean_var, log_std_var, step_log_std_var, step_hidden,
         hidden_init_var, dist) = model.build(self.input_var,
                                              self.step_input_var,
                                              step_hidden_var)

        # output layer is a tf.keras.layers.Dense object,
        # which cannot be access by tf.compat.v1.variable_scope.
        # A workaround is to access in tf.compat.v1.global_variables()
        for var in tf.compat.v1.global_variables():
            if 'output_layer/kernel' in var.name:
                std_share_output_weights = var
            if 'output_layer/bias' in var.name:
                std_share_output_bias = var
        assert std_share_output_weights.shape[1] == output_dim * 2
        assert std_share_output_bias.shape == output_dim * 2

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim, init_std', [
        (1, 1, 1),
        (1, 1, 2),
        (1, 2, 1),
        (1, 2, 2),
        (3, 3, 1),
        (3, 3, 2),
    ])
    # yapf: enable
    @mock.patch('tensorflow.random.normal')
    def test_without_std_share_network_output_values(self, mock_normal,
                                                     output_dim, hidden_dim,
                                                     init_std):
        mock_normal.return_value = 0.5
        model = GaussianGRUModel(output_dim=output_dim,
                                 hidden_dim=hidden_dim,
                                 std_share_network=False,
                                 hidden_nonlinearity=None,
                                 recurrent_nonlinearity=None,
                                 hidden_w_init=self.default_initializer,
                                 recurrent_w_init=self.default_initializer,
                                 output_w_init=self.default_initializer,
                                 init_std=init_std)
        step_hidden_var = tf.compat.v1.placeholder(shape=(self.batch_size,
                                                          hidden_dim),
                                                   name='step_hidden',
                                                   dtype=tf.float32)
        (mean_var, step_mean_var, log_std_var, step_log_std_var, step_hidden,
         hidden_init_var, dist) = model.build(self.input_var,
                                              self.step_input_var,
                                              step_hidden_var)

        hidden1 = hidden2 = np.full((self.batch_size, hidden_dim),
                                    hidden_init_var.eval())

        mean, log_std = self.sess.run(
            [mean_var, log_std_var],
            feed_dict={self.input_var: self.obs_inputs})

        for i in range(self.time_step):
            mean1, log_std1, hidden1 = self.sess.run(
                [step_mean_var, step_log_std_var, step_hidden],
                feed_dict={
                    self.step_input_var: self.obs_input,
                    step_hidden_var: hidden1
                })

            hidden2 = recurrent_step_gru(input_val=self.obs_input,
                                         num_units=hidden_dim,
                                         step_hidden=hidden2,
                                         w_x_init=0.1,
                                         w_h_init=0.1,
                                         b_init=0.,
                                         nonlinearity=None,
                                         gate_nonlinearity=None)

            output_nonlinearity = np.full(
                (np.prod(hidden2.shape[1:]), output_dim), 0.1)
            output2 = np.matmul(hidden2, output_nonlinearity)
            assert np.allclose(mean1, output2)
            expected_log_std = np.full((self.batch_size, output_dim),
                                       np.log(init_std))
            assert np.allclose(log_std1, expected_log_std)
            assert np.allclose(hidden1, hidden2)

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    def test_without_std_share_network_shapes(self, output_dim, hidden_dim):
        model = GaussianGRUModel(output_dim=output_dim,
                                 hidden_dim=hidden_dim,
                                 std_share_network=False,
                                 hidden_nonlinearity=None,
                                 recurrent_nonlinearity=None,
                                 hidden_w_init=self.default_initializer,
                                 recurrent_w_init=self.default_initializer,
                                 output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(shape=(self.batch_size,
                                                          hidden_dim),
                                                   name='step_hidden',
                                                   dtype=tf.float32)
        (mean_var, step_mean_var, log_std_var, step_log_std_var, step_hidden,
         hidden_init_var, dist) = model.build(self.input_var,
                                              self.step_input_var,
                                              step_hidden_var)

        # output layer is a tf.keras.layers.Dense object,
        # which cannot be access by tf.compat.v1.variable_scope.
        # A workaround is to access in tf.compat.v1.global_variables()
        for var in tf.compat.v1.global_variables():
            if 'output_layer/kernel' in var.name:
                std_share_output_weights = var
            if 'output_layer/bias' in var.name:
                std_share_output_bias = var
            if 'log_std_param/parameter' in var.name:
                log_std_param = var
        assert std_share_output_weights.shape[1] == output_dim
        assert std_share_output_bias.shape == output_dim
        assert log_std_param.shape == output_dim

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    @mock.patch('tensorflow.random.normal')
    def test_std_share_network_is_pickleable(self, mock_normal, output_dim,
                                             hidden_dim):
        mock_normal.return_value = 0.5
        model = GaussianGRUModel(output_dim=output_dim,
                                 hidden_dim=hidden_dim,
                                 std_share_network=True,
                                 hidden_nonlinearity=None,
                                 recurrent_nonlinearity=None,
                                 hidden_w_init=self.default_initializer,
                                 recurrent_w_init=self.default_initializer,
                                 output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(shape=(self.batch_size,
                                                          hidden_dim),
                                                   name='step_hidden',
                                                   dtype=tf.float32)
        (mean_var, step_mean_var, log_std_var, step_log_std_var, step_hidden,
         _, _) = model.build(self.input_var, self.step_input_var,
                             step_hidden_var)

        # output layer is a tf.keras.layers.Dense object,
        # which cannot be access by tf.compat.v1.variable_scope.
        # A workaround is to access in tf.compat.v1.global_variables()
        for var in tf.compat.v1.global_variables():
            if 'output_layer/bias' in var.name:
                var.load(tf.ones_like(var).eval())

        hidden = np.zeros((self.batch_size, hidden_dim))

        outputs1 = self.sess.run([mean_var, log_std_var],
                                 feed_dict={self.input_var: self.obs_inputs})
        output1 = self.sess.run([step_mean_var, step_log_std_var, step_hidden],
                                feed_dict={
                                    self.step_input_var: self.obs_input,
                                    step_hidden_var: hidden
                                })  # noqa: E126

        h = pickle.dumps(model)
        with tf.compat.v1.Session(graph=tf.Graph()) as sess:
            model_pickled = pickle.loads(h)

            input_var = tf.compat.v1.placeholder(tf.float32,
                                                 shape=(None, None,
                                                        self.feature_shape),
                                                 name='input')
            step_input_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=(None, self.feature_shape),
                name='step_input')
            step_hidden_var = tf.compat.v1.placeholder(shape=(self.batch_size,
                                                              hidden_dim),
                                                       name='initial_hidden',
                                                       dtype=tf.float32)

            (mean_var2, step_mean_var2, log_std_var2, step_log_std_var2,
             step_hidden2, _, _) = model_pickled.build(input_var,
                                                       step_input_var,
                                                       step_hidden_var)

            outputs2 = sess.run([mean_var2, log_std_var2],
                                feed_dict={input_var: self.obs_inputs})
            output2 = sess.run(
                [step_mean_var2, step_log_std_var2, step_hidden2],
                feed_dict={
                    step_input_var: self.obs_input,
                    step_hidden_var: hidden
                })
            assert np.array_equal(outputs1, outputs2)
            assert np.array_equal(output1, output2)

    # yapf: disable
    @pytest.mark.parametrize('output_dim, hidden_dim', [
        (1, 1),
        (2, 2),
        (3, 3)
    ])
    # yapf: enable
    @mock.patch('tensorflow.random.normal')
    def test_without_std_share_network_is_pickleable(self, mock_normal,
                                                     output_dim, hidden_dim):
        mock_normal.return_value = 0.5
        model = GaussianGRUModel(output_dim=output_dim,
                                 hidden_dim=hidden_dim,
                                 std_share_network=False,
                                 hidden_nonlinearity=None,
                                 recurrent_nonlinearity=None,
                                 hidden_w_init=self.default_initializer,
                                 recurrent_w_init=self.default_initializer,
                                 output_w_init=self.default_initializer)
        step_hidden_var = tf.compat.v1.placeholder(shape=(self.batch_size,
                                                          hidden_dim),
                                                   name='step_hidden',
                                                   dtype=tf.float32)
        (mean_var, step_mean_var, log_std_var, step_log_std_var, step_hidden,
         _, _) = model.build(self.input_var, self.step_input_var,
                             step_hidden_var)

        # output layer is a tf.keras.layers.Dense object,
        # which cannot be access by tf.compat.v1.variable_scope.
        # A workaround is to access in tf.compat.v1.global_variables()
        for var in tf.compat.v1.global_variables():
            if 'output_layer/bias' in var.name:
                var.load(tf.ones_like(var).eval())

        hidden = np.zeros((self.batch_size, hidden_dim))

        outputs1 = self.sess.run([mean_var, log_std_var],
                                 feed_dict={self.input_var: self.obs_inputs})
        output1 = self.sess.run([step_mean_var, step_log_std_var, step_hidden],
                                feed_dict={
                                    self.step_input_var: self.obs_input,
                                    step_hidden_var: hidden
                                })  # noqa: E126

        h = pickle.dumps(model)
        with tf.compat.v1.Session(graph=tf.Graph()) as sess:
            model_pickled = pickle.loads(h)

            input_var = tf.compat.v1.placeholder(tf.float32,
                                                 shape=(None, None,
                                                        self.feature_shape),
                                                 name='input')
            step_input_var = tf.compat.v1.placeholder(
                tf.float32,
                shape=(None, self.feature_shape),
                name='step_input')
            step_hidden_var = tf.compat.v1.placeholder(shape=(self.batch_size,
                                                              hidden_dim),
                                                       name='initial_hidden',
                                                       dtype=tf.float32)

            (mean_var2, step_mean_var2, log_std_var2, step_log_std_var2,
             step_hidden2, _, _) = model_pickled.build(input_var,
                                                       step_input_var,
                                                       step_hidden_var)

            outputs2 = sess.run([mean_var2, log_std_var2],
                                feed_dict={input_var: self.obs_inputs})
            output2 = sess.run(
                [step_mean_var2, step_log_std_var2, step_hidden2],
                feed_dict={
                    step_input_var: self.obs_input,
                    step_hidden_var: hidden
                })
            assert np.array_equal(outputs1, outputs2)
            assert np.array_equal(output1, output2)



import csv
from glob import glob
import json
import os
import os.path as osp

import joblib
import matplotlib.pyplot as plt
import numpy as np


def plot_experiments(name_or_patterns,
                     legend=False,
                     post_processing=None,
                     key='AverageReturn'):
    if not isinstance(name_or_patterns, (list, tuple)):
        name_or_patterns = [name_or_patterns]
    data_folder = osp.abspath(osp.join(osp.dirname(__file__), '../../data'))
    files = []
    for name_or_pattern in name_or_patterns:
        matched_files = glob(osp.join(data_folder, name_or_pattern))
        files += matched_files
    files = sorted(files)
    print('plotting the following experiments:')
    for f in files:
        print(f)
    plots = []
    legends = []
    for f in files:
        exp_name = osp.basename(f)
        returns = []
        with open(osp.join(f, 'progress.csv'), 'rb') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                if row[key]:
                    returns.append(float(row[key]))
        returns = np.array(returns)
        if post_processing:
            returns = post_processing(returns)
        plots.append(plt.plot(returns)[0])
        legends.append(exp_name)
    if legend:
        plt.legend(plots, legends)


class Experiment:
    def __init__(self, progress, params, pkl_data=None):
        self.progress = progress
        self.params = params
        self.pkl_data = pkl_data
        self.flat_params = self._flatten_params(params)
        self.name = params['exp_name']

    def _flatten_params(self, params, depth=2):
        flat_params = dict()
        for k, v in params.items():
            if isinstance(v, dict) and depth != 0:
                for subk, subv in self._flatten_params(
                        v, depth=depth - 1).items():
                    if subk == '_name':
                        flat_params[k] = subv
                    else:
                        flat_params[k + '_' + subk] = subv
            else:
                flat_params[k] = v
        return flat_params


def uniq(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


class ExperimentDatabase:
    def __init__(self, data_folder, names_or_patterns='*'):
        self._load_experiments(data_folder, names_or_patterns)

    def _read_data(self, progress_file):
        entries = dict()
        with open(progress_file, 'rb') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                for k, v in row.items():
                    if k not in entries:
                        entries[k] = []
                    entries[k].append(float(v))
        entries = dict([(k, np.array(v)) for k, v in entries.items()])
        return entries

    def _read_params(self, params_file):
        with open(params_file, 'r') as f:
            return json.loads(f.read())

    def _load_experiments(self, data_folder, name_or_patterns):
        if not isinstance(name_or_patterns, (list, tuple)):
            name_or_patterns = [name_or_patterns]
        files = []
        for name_or_pattern in name_or_patterns:
            matched_files = glob(
                osp.join(data_folder, name_or_pattern)
            )  # golb gives a list of all files satisfying pattern
            # this will include twice the same file if it satisfies 2 patterns
            files += matched_files
        experiments = []
        progress_f = None
        params_f = None
        pkl_data = None
        for f in files:
            if os.path.isdir(f):
                try:
                    progress = self._read_data(osp.join(f, 'progress.csv'))
                    params = self._read_params(osp.join(f, 'params.json'))
                    params['exp_name'] = osp.basename(f)
                    if os.path.isfile(osp.join(f, 'params.pkl')):
                        pkl_data = joblib.load(osp.join(f, 'params.pkl'))
                        experiments.append(
                            Experiment(progress, params, pkl_data))
                    else:
                        experiments.append(Experiment(progress, params))
                except Exception as e:
                    print(e)
            elif 'progress.csv' in f:
                # in case you're giving as datafolder the dir that contains
                # the files!
                progress_f = self._read_data(f)
            elif 'params.json' in f:
                params_f = self._read_params(f)
            elif 'params.pkl' in f:
                print('about to load', f)
                pkl_data = joblib.load(f)
        if params_f and progress_f:
            if pkl_data:
                experiments.append(Experiment(progress_f, params_f, pkl_data))
            else:
                experiments.append(Experiment(progress_f, params_f))

        self._experiments = experiments

    def plot_experiments(self,
                         key=None,
                         legend=None,
                         color_key=None,
                         filter_exp=None,
                         **kwargs):
        experiments = list(self.filter_experiments(**kwargs))
        if filter_exp:
            experiments = list(filter(filter_exp, experiments))
        plots = []
        legends = []
        color_pool = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w']
        color_map = dict()
        if color_key is not None:
            exp_color_keys = uniq(
                [exp.flat_params.get(color_key, None) for exp in experiments])
            if len(exp_color_keys) > len(color_pool):
                raise NotImplementedError
            for exp_color_key, color in zip(exp_color_keys, color_pool):
                print('%s: %s' % (str(exp_color_key), color))
            color_map = dict(list(zip(exp_color_keys, color_pool)))
        used_legends = []
        legend_list = []

        for exp in experiments:
            exp_color_key = None
            if color_key is not None:
                exp_color_key = exp.flat_params.get(color_key, None)
                exp_color = color_map.get(exp_color_key, None)
            else:
                exp_color = None
            plots.append(
                plt.plot(exp.progress.get(key, [0]), color=exp_color)[0])
            if legend is not None:
                legends.append(exp.flat_params[legend])
            elif (exp_color_key is not None
                  and exp_color_key not in used_legends):
                used_legends.append(exp_color_key)
                legend_list.append(plots[-1])

        if legends:
            plt.legend(plots, legends)
        elif legend_list:
            plt.legend(legend_list, used_legends)

    def filter_experiments(self, **kwargs):
        for exp in self._experiments:
            exp_params = exp.flat_params
            match = True
            for key, val in kwargs.items():
                if exp_params.get(key, None) != val:
                    match = False
                    break
            if match:
                yield exp

    def unique(self, param_key):
        return uniq([
            exp.flat_params[param_key] for exp in self._experiments
            if param_key in exp.flat_params
        ])



"""
This public package contains the replay buffer primitives.

The replay buffer primitives can be used for RL algorithms.
"""
from garage.replay_buffer.her_replay_buffer import HerReplayBuffer
from garage.replay_buffer.path_buffer import PathBuffer
from garage.replay_buffer.simple_replay_buffer import SimpleReplayBuffer

__all__ = ['HerReplayBuffer', 'PathBuffer', 'SimpleReplayBuffer']



'''
This script creates a regression test over garage-PPO and baselines-PPO.
Unlike garage, baselines doesn't set max_path_length. It keeps steps the action
until it's done. So we introduced tests.wrappers.AutoStopEnv wrapper to set
done=True when it reaches max_path_length.
'''
import datetime
import os.path as osp
import random

import dowel
from dowel import logger as dowel_logger
import gym
import pytest
import tensorflow as tf

from garage.envs import normalize
from garage.experiment import deterministic
from garage.np.baselines import LinearFeatureBaseline
from garage.tf.algos import PPO
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import CategoricalMLPPolicy
from tests.fixtures import snapshot_config
import tests.helpers as Rh


class TestBenchmarkCategoricalMLPPolicy:
    '''Compare benchmarks between garage and baselines.'''

    @pytest.mark.huge
    def test_benchmark_categorical_mlp_policy(self):
        '''
        Compare benchmarks between garage and baselines.
        :return:
        '''
        categorical_tasks = [
            'LunarLander-v2', 'CartPole-v1', 'Assault-ramDeterministic-v4',
            'Breakout-ramDeterministic-v4',
            'ChopperCommand-ramDeterministic-v4',
            'Tutankham-ramDeterministic-v4'
        ]
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S-%f')
        benchmark_dir = './data/local/benchmarks/categorical_mlp_policy/{0}/'
        benchmark_dir = benchmark_dir.format(timestamp)
        result_json = {}
        for task in categorical_tasks:
            env_id = task
            env = gym.make(env_id)
            trials = 3
            seeds = random.sample(range(100), trials)

            task_dir = osp.join(benchmark_dir, env_id)
            plt_file = osp.join(benchmark_dir,
                                '{}_benchmark.png'.format(env_id))
            relplt_file = osp.join(benchmark_dir,
                                   '{}_benchmark_mean.png'.format(env_id))
            garage_csvs = []

            for trial in range(trials):
                seed = seeds[trial]

                trial_dir = task_dir + '/trial_%d_seed_%d' % (trial + 1, seed)
                garage_dir = trial_dir + '/garage'

                with tf.Graph().as_default():
                    # Run garage algorithms
                    env.reset()
                    garage_csv = run_garage(env, seed, garage_dir)
                garage_csvs.append(garage_csv)

            env.close()

            Rh.plot(b_csvs=garage_csvs,
                    g_csvs=garage_csvs,
                    g_x='Iteration',
                    g_y='AverageReturn',
                    g_z='Garage',
                    b_x='Iteration',
                    b_y='AverageReturn',
                    b_z='Garage',
                    trials=trials,
                    seeds=seeds,
                    plt_file=plt_file,
                    env_id=env_id,
                    x_label='Iteration',
                    y_label='AverageReturn')

            Rh.relplot(b_csvs=garage_csvs,
                       g_csvs=garage_csvs,
                       g_x='Iteration',
                       g_y='AverageReturn',
                       g_z='Garage',
                       b_x='Iteration',
                       b_y='AverageReturn',
                       b_z='Garage',
                       trials=trials,
                       seeds=seeds,
                       plt_file=relplt_file,
                       env_id=env_id,
                       x_label='Iteration',
                       y_label='AverageReturn')

            result_json[env_id] = Rh.create_json(b_csvs=garage_csvs,
                                                 g_csvs=garage_csvs,
                                                 seeds=seeds,
                                                 trails=trials,
                                                 g_x='Iteration',
                                                 g_y='AverageReturn',
                                                 b_x='Iteration',
                                                 b_y='AverageReturn',
                                                 factor_g=2048,
                                                 factor_b=2048)

        Rh.write_file(result_json, 'PPO')


def run_garage(env, seed, log_dir):
    '''
    Create garage model and training.
    :param env: Environment of the task.
    :param seed: Random seed for the trial.
    :param log_dir: Log dir path.
    :return:
    '''
    deterministic.set_seed(seed)
    config = tf.ConfigProto(allow_soft_placement=True,
                            intra_op_parallelism_threads=12,
                            inter_op_parallelism_threads=12)
    sess = tf.Session(config=config)
    with LocalTFRunner(snapshot_config, sess=sess, max_cpus=12) as runner:
        env = TfEnv(normalize(env))

        policy = CategoricalMLPPolicy(
            env_spec=env.spec,
            hidden_nonlinearity=tf.nn.tanh,
        )

        baseline = LinearFeatureBaseline(env_spec=env.spec)

        algo = PPO(env_spec=env.spec,
                   policy=policy,
                   baseline=baseline,
                   max_path_length=100,
                   discount=0.99,
                   gae_lambda=0.95,
                   lr_clip_range=0.2,
                   policy_ent_coeff=0.0,
                   optimizer_args=dict(
                       batch_size=32,
                       max_epochs=10,
                       tf_optimizer_args=dict(learning_rate=1e-3),
                   ),
                   name='CategoricalMLPPolicyBenchmark')

        # Set up logger since we are not using run_experiment
        tabular_log_file = osp.join(log_dir, 'progress.csv')
        dowel_logger.add_output(dowel.StdOutput())
        dowel_logger.add_output(dowel.CsvOutput(tabular_log_file))
        dowel_logger.add_output(dowel.TensorBoardOutput(log_dir))

        runner.setup(algo, env, sampler_args=dict(n_envs=12))
        runner.train(n_epochs=5, batch_size=2048)
        dowel_logger.remove_all()

        return tabular_log_file



import numpy as np
import tensorflow as tf
from garage.tf.distributions.diagonal_gaussian import DiagonalGaussian
from garage.tf.policies.base import StochasticPolicy


class GoExplorePolicy(StochasticPolicy):
    """A stochastic policy for Go-Explore that takes actions uniformally at random.

    Parameters
    ----------
    env_spec : :py:class:`garage.envs.EnvSpec`
        Environment specification.
    name : str
        Name for the tensors.
    """

    def __init__(self, env_spec, name='GoExplorePolicy'):

        self.dist = DiagonalGaussian(dim=env_spec.action_space.flat_dim)
        self.log_std = np.zeros(env_spec.action_space.flat_dim)

        super(GoExplorePolicy, self).__init__(env_spec=env_spec, name=name)
        self._initialize()

    def _initialize(self):
        """Initialize the tensor variable scope.

        """
        with tf.compat.v1.variable_scope(self.name) as vs:
            self._variable_scope = vs

    # Should be implemented by all policies
    def get_action(self, observation):
        """Get action sampled from the policy.

        Parameters
        ----------
        observation : array_like
            Observation from the environment.

        Returns
        -------
        array_like
            Action sampled from the policy.

        """
        return self.action_space.sample(), dict(mean=self.log_std, log_std=self.log_std)

    def get_actions(self, observations):
        """Get actions sampled from the policy.

        Parameters
        ----------
        observations : list[array_like]
            Observations from the environment.

        Returns
        -------
        array_like
            Actions sampled from the policy.

        """
        means = [self.log_std for observation in observations]
        log_stds = [self.log_std for observation in observations]
        return self.action_space.sample_n(len(observations)), dict(mean=means, log_std=log_stds)

    # def get_params_internal(self, **tags):
    #     """
    #
    #     Parameters
    #     ----------
    #     tags :
    #
    #     Returns
    #     -------
    #
    #     """
    #     return []

    def reset(self, dones=None):
        """Reset the policy.

        If dones is None, it will be by default np.array([True]) which implies
        the policy will not be "vectorized", i.e. number of parallel
        environments for training data sampling = 1.

        Parameters
        ----------
        dones : array_like
            Bool that indicates terminal state(s).
        """

    def log_diagnostics(self, paths):
        """
        Log extra information per iteration based on the collected paths.
        """

    @property
    def vectorized(self):
        """
        Indicates whether the policy is vectorized. If True, it should
        implement get_actions(), and support resetting
        with multiple simultaneous states.
        """
        return False

    def terminate(self):
        """
        Clean up operation.
        """

    @property
    def distribution(self):
        """Distribution.

        Returns
        -------

            Distribution.
        """
        return self.dist

    def dist_info(self, obs, state_infos):
        """
        Distribution info.

        Return the distribution information about the actions.

        Parameters
        ----------
        obs : array_like
             Observation values.
        state_infos : dict
            A dictionary whose values should contain
            information about the state of the policy at the time it received the
            observation.
        """
        return dict(mean=None, log_std=self.log_std)

    def dist_info_sym(self, obs_var, state_info_vars, name='dist_info_sym'):
        """Symbolic graph of the distribution.

        Return the symbolic distribution information about the actions.

        Parameters
        ----------
        obs_var : `tf.Tensor <https://www.tensorflow.org/api_docs/python/tf/Tensor>`_
            Symbolic variable for observations.
        state_infos : dict
            A dictionary whose values should contain
            information about the state of the policy at the time it received the
            observation.
        name : str
            Name of the symbolic graph.
        """
        raise NotImplementedError

    # def get_param_values(self):
    #     print("params ", self.cell_num, ", ", self.stateful_num, ", ", self.cell_pool, " retrieved from ", self)
    #     return {"cell_num": self.cell_num,
    #             "stateful_num": self.stateful_num,
    #             "cell_pool": self.cell_pool}
    #
    # def set_param_values(self, params):
    #     self.cell_num = params["cell_num"]
    #     self.stateful_num = params["stateful_num"]
    #     self.cell_pool = params["cell_pool"]
    #     print(self, " had params set to ", self.cell_num, ", ", self.stateful_num)
    #
    # def set_cell_pool(self, cell_pool):
    #     self.cell_pool = cell_pool
    #     print(self, "had cell pool set to: ", self.cell_pool)



import numpy as np


def log_mean_exp(x, dim):
    """
    Compute the log(mean(exp(x), dim)) in a numerically stable manner
    """
    return log_sum_exp(x, dim) - np.log(x.shape[dim])


def log_sum_exp(x, dim):
    """
    Compute the log(sum(exp(x), dim)) in a numerically stable manner
    """
    max_x = np.max(x, dim)
    new_x = x - np.repeat(np.expand_dims(max_x, dim), x.shape[dim], dim)
    return max_x + np.log(np.sum(np.exp(new_x), dim))


def softmax(x, dim):
    """Compute softmax values for each sets of scores in x along dim"""
    e_x = np.exp(x - np.max(x, dim))
    return e_x / np.sum(e_x, dim)



"""
This script creates a test that fails when
garage.tf.baselines failed to initialize.
"""
import tensorflow as tf

from garage.tf.baselines import ContinuousMLPBaseline
from garage.tf.baselines import GaussianMLPBaseline
from garage.tf.envs import TfEnv
from tests.fixtures import TfGraphTestCase
from tests.fixtures.envs.dummy import DummyBoxEnv


class TestTfBaselines(TfGraphTestCase):

    def test_baseline(self):
        """Test the baseline initialization."""
        box_env = TfEnv(DummyBoxEnv())
        deterministic_mlp_baseline = ContinuousMLPBaseline(env_spec=box_env)
        gaussian_mlp_baseline = GaussianMLPBaseline(env_spec=box_env)

        self.sess.run(tf.compat.v1.global_variables_initializer())
        deterministic_mlp_baseline.get_param_values(trainable=True)
        gaussian_mlp_baseline.get_param_values(trainable=True)

        box_env.close()



#!/usr/bin/env python3
"""This is an example to train a task with DDPG algorithm written in PyTorch.

Here it creates a gym environment InvertedDoublePendulum. And uses a DDPG with
1M steps.

"""
import gym
import torch
from torch.nn import functional as F  # NOQA

from garage.envs import normalize
from garage.envs.base import GarageEnv
from garage.experiment import LocalRunner, run_experiment
from garage.np.exploration_strategies import OUStrategy
from garage.replay_buffer import SimpleReplayBuffer
from garage.torch.algos import DDPG
from garage.torch.policies import DeterministicMLPPolicy
from garage.torch.q_functions import ContinuousMLPQFunction


def run_task(snapshot_config, *_):
    """Set up environment and algorithm and run the task."""
    runner = LocalRunner(snapshot_config)
    env = GarageEnv(normalize(gym.make('InvertedDoublePendulum-v2')))

    action_noise = OUStrategy(env.spec, sigma=0.2)

    policy = DeterministicMLPPolicy(env_spec=env.spec,
                                    hidden_sizes=[64, 64],
                                    hidden_nonlinearity=F.relu,
                                    output_nonlinearity=torch.tanh)

    qf = ContinuousMLPQFunction(env_spec=env.spec,
                                hidden_sizes=[64, 64],
                                hidden_nonlinearity=F.relu)

    replay_buffer = SimpleReplayBuffer(env_spec=env.spec,
                                       size_in_transitions=int(1e6),
                                       time_horizon=100)

    ddpg = DDPG(env_spec=env.spec,
                policy=policy,
                qf=qf,
                replay_buffer=replay_buffer,
                n_train_steps=50,
                min_buffer_size=int(1e4),
                exploration_strategy=action_noise,
                target_update_tau=1e-2,
                policy_lr=1e-4,
                qf_lr=1e-3,
                discount=0.9,
                optimizer=torch.optim.Adam)

    runner.setup(algo=ddpg, env=env)

    runner.train(n_epochs=500, n_epoch_cycles=20, batch_size=100)


run_experiment(
    run_task,
    snapshot_mode='last',
    seed=1,
)



import numpy as np
import tensorflow as tf
from garage.misc.tensor_utils import flatten_tensors
from garage.misc.tensor_utils import unflatten_tensors


# initializ parameter with numpy (easier for seeding)
# bias initialzied with0, weight initialzied with XavierUniformInitializer
def init_param_np(param, policy, np_random=np.random):
    assert param.name[-3] == "W" or param.name[-3] == "b"
    if param.name[-3] == "W":
        shape = tf.shape(param).eval()
        if len(shape) == 2:
            n_inputs, n_outputs = shape
        else:
            receptive_field_size = np.prod(shape[:2])
            n_inputs = shape[-2] * receptive_field_size
            n_outputs = shape[-1] * receptive_field_size
        init_range = np.sqrt(6.0 / (n_inputs + n_outputs))
        # return tf.random_uniform_initializer(-init_range, init_range, dtype=dtype)(shape)
        param_value = (np_random.rand(*shape) * 2 - 1) * init_range
    elif param.name[-3] == "b":
        param_value = np.zeros_like(param.eval())
    feed_dict = dict()
    feed_dict[policy._cached_assign_placeholders[param]] = param_value
    tf.get_default_session().run(policy._cached_assign_ops[param], feed_dict=feed_dict)


def init_policy_np(policy, np_random=np.random):
    params = policy.get_params(trainable=True)
    shapes = policy.get_param_shapes(trainable=True)
    param_values = policy.get_param_values(trainable=True)

    flattened_params = np_random.rand(*param_values.shape)
    param_values = unflatten_tensors(flattened_params, shapes)

    for i, param in enumerate(params):
        # assert param.name[-3] == "W" or param.name[-3] == "b"
        if param.name[-3] == "W":
            shape = shapes[i]
            if len(shape) == 2:
                n_inputs, n_outputs = shape
            else:
                receptive_field_size = np.prod(shape[:2])
                n_inputs = shape[-2] * receptive_field_size
                n_outputs = shape[-1] * receptive_field_size
            init_range = np.sqrt(6.0 / (n_inputs + n_outputs))
            param_values[i] = (param_values[i] * 2 - 1) * init_range
        elif param.name[-3] == "b":
            param_values[i] = np.zeros_like(param_values[i])

    param_values = flatten_tensors(param_values)
    return param_values


# class XavierUniformInitializer(object):
#     def __call__(self, shape, dtype=tf.float32, *args, **kwargs):
#         if len(shape) == 2:
#             n_inputs, n_outputs = shape
#         else:
#             receptive_field_size = np.prod(shape[:2])
#             n_inputs = shape[-2] * receptive_field_size
#             n_outputs = shape[-1] * receptive_field_size
#         init_range = math.sqrt(6.0 / (n_inputs + n_outputs))
#         return tf.random_uniform_initializer(-init_range, init_range, dtype=dtype)(shape)


# class HeUniformInitializer(object):
#     def __call__(self, shape, dtype=tf.float32, *args, **kwargs):
#         if len(shape) == 2:
#             n_inputs, _ = shape
#         else:
#             receptive_field_size = np.prod(shape[:2])
#             n_inputs = shape[-2] * receptive_field_size
#         init_range = math.sqrt(1.0 / n_inputs)
#         return tf.random_uniform_initializer(-init_range, init_range, dtype=dtype)(shape)



import abc


class Sampler(abc.ABC):
    @abc.abstractmethod
    def start_worker(self):
        """Initialize the sampler,

        e.g. launching parallel workers if necessary.
        """
        pass

    @abc.abstractmethod
    def obtain_samples(self, itr):
        """Collect samples for the given iteration number.

        Args:
            itr (int): Iteration number.

        Returns:
            list[dict]: A list of paths.

        """
        pass

    @abc.abstractmethod
    def shutdown_worker(self):
        """Terminate workers if necessary."""
        pass


class BaseSampler(Sampler):
    """Base class for sampler.

    Args:
        algo (garage.np.algos.RLAlgorithm): The algorithm.
        env (gym.Env): The environment.

    """

    def __init__(self, algo, env):
        self.algo = algo
        self.env = env



"""Algorithms for solving AST formulated RL problems."""
import importlib

from .ga import GA  # noqa
from .gasm import GASM  # noqa
from .mcts import MCTS  # noqa
from .mctsbv import MCTSBV  # noqa
from .mctsrs import MCTSRS  # noqa

__all__ = ['GA', 'GASM', 'MCTS', 'MCTSBV', 'MCTSRS']

if importlib.util.find_spec('bsddb3') is not None:
    # Only load bsddb3 dependent packages if it is installed
    from .go_explore import GoExplore  # noqa
    from .backward_algorithm import BackwardAlgorithm  # noqa

    __all__ += ['GoExplore', 'BackwardAlgorithm']

# from .random_search import RandomSearch
# from .trpo import TRPO



from tests.fixtures.policies.dummy_policy import (DummyPolicy,
                                                  DummyPolicyWithoutVectorized)
from tests.fixtures.policies.dummy_recurrent_policy import (
    DummyRecurrentPolicy)

__all__ = [
    'DummyPolicy', 'DummyRecurrentPolicy', 'DummyPolicyWithoutVectorized'
]



from unittest import mock

import gym
import numpy as np

from tests.fixtures.envs.dummy import DummyEnv


class DummyDiscretePixelEnv(DummyEnv):
    """
    A dummy discrete pixel environment.

    It follows Atari game convention, where actions are 'NOOP', 'FIRE', ...
    It also contains self.unwrapped.ale.lives, get_action_meanings for testing.

    Several properties are made for testing purpose as following:

    -Observations are
        after reset    : np.ones(self._shape).
        action 1 (FIRE): np.full(self._shape, 2).
        otherwise      : random if self.random is True,
            otherwise previous state + action.

    -The environment has 5 lives.
    -Done will be True if
        -all 5 lives are exhausted
        -env.step(2), followed by env.step(1)
    """

    def __init__(self, random=True):
        super().__init__(random, obs_dim=(10, 10, 3), action_dim=5)
        self.unwrapped.get_action_meanings = self._get_action_meanings
        self.unwrapped.ale = mock.Mock()
        self.unwrapped.ale.lives = self.get_lives
        self._observation_space = gym.spaces.Box(
            low=0, high=255, shape=self._obs_dim, dtype=np.uint8)
        self.step_called = 0
        self._prev_action = None

    @property
    def observation_space(self):
        """Return an observation space."""
        return self._observation_space

    @observation_space.setter
    def observation_space(self, observation_space):
        self._observation_space = observation_space

    @property
    def action_space(self):
        """Return an action space."""
        return gym.spaces.Discrete(self._action_dim)

    def _get_action_meanings(self):
        return ['NOOP', 'FIRE', 'SLEEP', 'EAT', 'PLAY']

    def get_lives(self):
        """Get number of lives."""
        return self._lives

    def reset(self):
        """Reset the environment."""
        self.state = np.ones(self._obs_dim, dtype=np.uint8)
        self._lives = 5
        self.step_called = 0
        return self.state

    def step(self, action):
        """
        Step the environment.

        Before gym fixed overflow issue for sample() in
        np.uint8 environment, we will handle the sampling here.
        We need high=256 since np.random.uniform sample from [low, high)
        (includes low, but excludes high).
        """
        done = False
        if self.state is not None:
            # Simulating FIRE action
            if action == 1:
                if self._prev_action == 2:
                    done = True
                obs = np.full(self._obs_dim, 2, dtype=np.uint8)
            else:
                if self.random:
                    obs = np.random.uniform(
                        low=0, high=256, size=self._obs_dim).astype(np.uint8)
                else:
                    obs = self.state + action
            if self._lives == 0:
                raise RuntimeError("DummyEnv: Cannot step when lives = 0!")
            self._lives -= 1
            if self._lives == 0:
                done = True
        else:
            raise RuntimeError(
                "DummyEnv: reset() must be called before step()!")
        self.step_called += 1
        self._prev_action = action

        return obs, 0, done, {'ale.lives': self._lives}



import numpy as np
import tensorflow as tf

from garage.tf.regressors import StochasticRegressor
from tests.fixtures.models import SimpleGaussianMLPModel


class SimpleGaussianMLPRegressor(StochasticRegressor):
    """Simple GaussianMLPRegressor for testing."""

    def __init__(self, input_shape, output_dim, name, *args, **kwargs):
        super().__init__(input_shape, output_dim, name)

        self.model = SimpleGaussianMLPModel(output_dim=self._output_dim)

        self._ys = None
        self._initialize()

    def _initialize(self):
        input_ph = tf.compat.v1.placeholder(tf.float32,
                                            shape=(None, ) + self._input_shape)
        with tf.compat.v1.variable_scope(self._name) as vs:
            self._variable_scope = vs
            self.model.build(input_ph)

    def fit(self, xs, ys):
        self._ys = ys

    def predict(self, xs):
        if self._ys is None:
            mean = tf.compat.v1.get_default_session().run(
                self.model.networks['default'].mean,
                feed_dict={self.model.networks['default'].input: xs})
            self._ys = np.full((len(xs), 1), mean)

        return self._ys

    def get_params_internal(self, *args, **kwargs):
        return self._variable_scope.trainable_variables()

    def __setstate__(self, state):
        """Object.__setstate__."""
        super().__setstate__(state)
        self._initialize()



from unittest.mock import patch

import numpy as np
import pytest
from bsddb3 import db
from gym.spaces.box import Box

import ast_toolbox.samplers.parallel_sampler as ps
from ast_toolbox.algos.go_explore import Cell
from ast_toolbox.algos.go_explore import CellPool
from ast_toolbox.envs.go_explore_ast_env import Custom_GoExploreASTEnv
from ast_toolbox.envs.go_explore_ast_env import GoExploreASTEnv
from ast_toolbox.envs.go_explore_ast_env import GoExploreParameter
from ast_toolbox.simulators import ASTSimulator
from ast_toolbox.simulators import ExampleAVSimulator
from ast_toolbox.spaces import ASTSpaces
from ast_toolbox.spaces import ExampleAVSpaces
from examples.AV.example_runner_ba_av import runner as ba_runner
from examples.AV.example_runner_drl_av import runner as drl_runner
from examples.AV.example_runner_ga_av import runner as ga_runner
from examples.AV.example_runner_ge_av import runner as ge_runner
from examples.AV.example_runner_mcts_av import runner as mcts_runner
from tests.validate_drl import validate_drl
from tests.validate_ga import validate_ga
from tests.validate_ge_ba import validate_ge_ba
from tests.validate_install import validate_install
from tests.validate_mcts import validate_mcts
from tests.validate_parallel import validate_parallel


def test_validate_install():
    assert validate_install() is True


def test_validate_parallel():
    assert validate_parallel() is True


def test_validate_drl():
    assert validate_drl() is True


def test_validate_mcts():
    assert validate_mcts() is True


def test_validate_ga():
    assert validate_ga() is True


def test_validate_ge_ba():
    assert validate_ge_ba() is True


def test_ast_simulator():
    sim = ASTSimulator()
    with pytest.raises(NotImplementedError):
        sim.simulate(None, None)
    sim.open_loop = False
    with pytest.raises(NotImplementedError):
        sim.get_reward_info()
    with pytest.raises(NotImplementedError):
        sim.is_goal()

    sim.open_loop = True
    sim.initial_conditions = np.array([0, 0, 0, 0, 0])
    sim.observation = np.array([1, 1, 1, 1])

    sim.c_max_path_length = 2
    sim._path_length = 0

    assert np.all(sim.step(None) == np.array([0, 0, 0, 0, 0]))
    assert (sim.is_terminal() is False)

    assert np.all(sim.step(None) == np.array([0, 0, 0, 0, 0]))
    assert (sim.is_terminal() is True)

    sim.blackbox_sim_state = False
    assert np.all(sim.observation_return() == np.array([1, 1, 1, 1]))
    assert np.all(sim.reset(np.array([0, 0, 0, 0, 0])) == np.array([1, 1, 1, 1]))
    sim.open_loop = False
    assert np.all(sim.closed_loop_step(action=np.array([])) == np.array([1, 1, 1, 1]))
    sim.open_loop = True

    sim.blackbox_sim_state = True
    assert np.all(sim.observation_return() == np.array([0, 0, 0, 0, 0]))
    assert np.all(sim.reset(np.array([0, 0, 0, 0, 0])) == np.array([0, 0, 0, 0, 0]))
    sim.open_loop = False
    assert np.all(sim.closed_loop_step(action=np.array([])) == np.array([0, 0, 0, 0, 0]))
    sim.open_loop = True

    assert sim.log() is None


def test_example_av_simulator():
    simulator_args = {'car_init_x': 0,
                      'car_init_y': 0}
    sim = ExampleAVSimulator(simulator_args=simulator_args, max_path_length=1)
    sim.blackbox_sim_state = False

    # check reset
    init = sim.reset(s_0=np.array([0, 0, 0, 0, 0]))
    assert np.all(init == np.array([0, 0, 0, 0]))

    # Check if simulate can find a goal
    path_length, info = sim.simulate(actions=[np.zeros(6)], s_0=np.array([0, 0, 0, 1, 0]))
    assert path_length == 0
    # Check if simulate can return end of path
    path_length, info = sim.simulate(actions=[np.zeros(6)], s_0=np.array([5, 5, 0, 0, 0]))
    assert path_length == -1

    # Check open-loop sim step
    sim.open_loop = True
    sim.reset(s_0=np.array([0, 0, 0, 0, 0]))
    obs = sim.step(action=np.zeros(6))
    assert np.all(obs == np.array([0, 0, 0, 0, 0]))

    # Check closed-loop sim step
    sim.open_loop = False
    init = sim.reset(s_0=np.array([0, 0, 0, 0, 0]))
    obs = sim.step(action=np.zeros(6))
    assert np.all(obs == np.array([0, 0, 0, 0]))


def test_ast_spaces():
    space = ASTSpaces()

    with pytest.raises(NotImplementedError):
        space.action_space()

    with pytest.raises(NotImplementedError):
        space.observation_space()


def test_example_av_spaces():
    space = ExampleAVSpaces(num_peds=2)

    assert isinstance(space.action_space, Box)
    assert isinstance(space.observation_space, Box)


def test_go_explore_ast_env():
    env = GoExploreASTEnv()
    env._fixed_init_state = True
    env._init_state = np.zeros(5)
    env.p_robustify_state = None
    # import pdb; pdb.set_trace()
    with patch('ast_toolbox.envs.go_explore_ast_env.db.DB', side_effect=db.DBBusyError):
        assert np.all(env.reset() == np.zeros(5))
    with patch('ast_toolbox.envs.go_explore_ast_env.db.DB', side_effect=db.DBLockNotGrantedError):
        assert np.all(env.reset() == np.zeros(5))
    with patch('ast_toolbox.envs.go_explore_ast_env.db.DB', side_effect=db.DBForeignConflictError):
        assert np.all(env.reset() == np.zeros(5))
    with patch('ast_toolbox.envs.go_explore_ast_env.db.DB', side_effect=db.DBAccessError):
        assert np.all(env.reset() == np.zeros(5))
    with patch('ast_toolbox.envs.go_explore_ast_env.db.DB', side_effect=db.DBPermissionsError):
        assert np.all(env.reset() == np.zeros(5))
    with patch('ast_toolbox.envs.go_explore_ast_env.db.DB', side_effect=db.DBNoSuchFileError):
        assert np.all(env.reset() == np.zeros(5))
    with patch('ast_toolbox.envs.go_explore_ast_env.db.DB', side_effect=db.DBError):
        assert np.all(env.reset() == np.zeros(5))

    env.p_key_list = GoExploreParameter(name='key_list', value=[0])
    env.p_max_value = GoExploreParameter(name='max_value', value=1)
    env.sample_limit = 10

    # env.p_key_list.value = [0]

    class Test_Pop:
        def __init__(self, fitness):
            self.fitness = fitness

    test_pop = Test_Pop(0.1)
    with patch('ast_toolbox.envs.go_explore_ast_env.random.random', return_value=1.0):
        assert env.sample(population=[test_pop]) == test_pop
    test_pop = Test_Pop(0)
    with patch('ast_toolbox.envs.go_explore_ast_env.random.random', return_value=1.0):
        with pytest.raises(ValueError):
            env.sample(population=[test_pop])

    env.simulator.blackbox_sim_state = True
    env.simulator.initial_conditions = np.zeros(5)
    # import pdb; pdb.set_trace()
    env.render(car=None, ped=None, noise=None)

    env._init_state = np.array([0, 0, 0, 1, 0])
    env.simulator.c_max_path_length = 1
    path_length, info = env.simulate(actions=[np.zeros(6)])
    assert path_length == 0
    env._fixed_init_state = False
    path_length, info = env.simulate(actions=[np.zeros(6)])
    assert path_length == -1

    env.blackbox_sim_state = True
    env.simulator.blackbox_sim_state = True
    obs = env.env_reset()
    assert env.observation_space.contains(obs)

    env.blackbox_sim_state = False
    env.simulator.blackbox_sim_state = False
    obs = env.env_reset()
    assert env.observation_space.contains(obs[4:])

    obs = np.array([0, 0.001, 0.0015, 0.002])
    assert np.all(env.downsample(obs) == obs)

    cenv = Custom_GoExploreASTEnv()
    cenv._step = 1
    assert np.all(cenv.downsample(obs) == np.array([1, 0, 1, 1, 2]))

    assert isinstance(env.observation_space, Box)
    assert isinstance(env.action_space, Box)

    env._info = 'test'
    assert env.get_cache_list() == 'test'


def test_parallel_sampler():
    # env = TfEnv(env_name='CartPole-v1')
    #
    # policy = CategoricalMLPPolicy(
    #     name='policy', env_spec=env.spec, hidden_sizes=(32, 32))
    env = None
    policy = None
    ps.initialize(1)
    ps.set_seed(0)
    ps.populate_task(env, policy)
    ps.close()

    ps.initialize(2)
    ps.set_seed(0)
    ps.populate_task(env, policy, scope=0)
    ps.populate_task(env, policy, scope=1)
    ps.close()


def test_go_explore():
    cell1 = Cell(use_score_weight=False)
    cell2 = Cell(use_score_weight=False)
    cell1.observation = np.zeros(5)
    cell2.observation = np.zeros(5)

    assert cell1 != np.zeros(5)
    assert cell1 == cell2

    cell2.observation = np.ones(5)
    assert cell1 != cell2

    assert cell1.is_root
    assert cell1.step == 0

    cell1.fitness
    assert 'fitness' in cell1.__dict__
    cell1.times_visited = 1
    assert 'fitness' not in cell1.__dict__

    assert hash(cell1) == hash((cell1.observation.tostring()))

    cell_pool = CellPool(filename='./test_pool.dat', use_score_weight=True)
    d_pool = cell_pool.open_pool(overwrite=True)
    cell_pool.d_update(
        cell_pool_shelf=d_pool,
        observation=np.zeros(5),
        action=np.zeros(5),
        trajectory=np.array(
            []),
        score=0.0,
        state=None,
        reward=0.0,
        chosen=0)
    cell_pool.d_update(cell_pool_shelf=d_pool, observation=np.zeros(5), action=np.zeros(5), trajectory=np.array([]), score=1.0,
                       state=None, reward=1.0, chosen=0, is_goal=True)


def test_example_runner_ba_av():
    sim = ExampleAVSimulator()
    sim.reset(np.zeros(5))
    expert_trajectory = {'state': sim.clone_state(),
                         'reward': 0,
                         'action': np.zeros(6),
                         'observation': np.zeros(5)}
    with patch('examples.AV.example_runner_ba_av.compress_pickle.dump', side_effect=MemoryError):
        with patch('examples.AV.example_runner_ba_av.LocalTFRunner.train', new=lambda x: 0):
            ba_runner(
                env_args={
                    'id': 'ast_toolbox:GoExploreAST-v1'},
                algo_args={
                    'expert_trajectory': [expert_trajectory] *
                    50,
                    'max_epochs': 10})


def test_example_runner_drl_av():
    with patch('examples.AV.example_runner_drl_av.run_experiment'):
        drl_runner(save_expert_trajectory=False)
#     # Create mock data from last iteration
#     mock_steps = 3
#     mock_env_info = {'actions':np.zeros((mock_steps, 5)),
#                  'state':np.zeros((mock_steps, 5)),}
#     mock_path = {'rewards':np.zeros(mock_steps),
#                  'observations':np.zeros((mock_steps, 5)),
#                  'env_infos':mock_env_info}
#     mock_paths = [mock_path]
#     mock_last_iter_data = {'paths':mock_paths}
#
#
#     with patch('examples.AV.test_example_runner_drl_av.LocalTFRunner.train'):
#         drl_runner(save_expert_trajectory=False)
#         with patch('examples.AV.test_example_runner_drl_av.open'):
#             with patch('examples.AV.test_example_runner_drl_av.pickle.load', return_value = mock_last_iter_data):
#                 with patch('examples.AV.test_example_runner_drl_av.pickle.dump'):
#                     drl_runner(save_expert_trajectory=True)


def test_example_runner_ga_av():
    with patch('examples.AV.example_runner_ga_av.run_experiment'):
        ga_runner()


def test_example_runner_ge_av():
    with patch('examples.AV.example_runner_ge_av.run_experiment'):
        ge_runner()


def test_example_runner_mcts_av():
    with patch('examples.AV.example_runner_mcts_av.run_experiment'):
        mcts_runner()



import numpy as np
from garage.sampler.on_policy_vectorized_sampler import OnPolicyVectorizedSampler

from ast_toolbox.rewards import ExampleAVReward
from ast_toolbox.simulators import ExampleAVSimulator


class ASTVectorizedSampler(OnPolicyVectorizedSampler):
    """A vectorized sampler for AST to handle open-loop simulators.

    Garage usually genearates samples in a closed-loop process. This version of the vectorized sampler instead
    grabs dummy data until the full rollout specification is generated, then goes back and runs the `simulate`
    function to actually obtain results. Rewards are then calculated and the path data is corrected.

    Parameters
    ----------
    algo : :py:class:`garage.np.algos.base.RLAlgorithm`
        The algorithm.
    env : :py:class:`ast_toolbox.envs.ASTEnv`
        The environment.
    n_envs : int
        Number of parallel environments to run.
    open_loop : bool
        True if the simulation is open-loop, meaning that AST must generate all actions ahead of time, instead
        of being able to output an action in sync with the simulator, getting an observation back before
        the next action is generated. False to get interactive control, which requires that `blackbox_sim_state`
        is also False.
    sim : :py:class:`ast_toolbox.simulators.ASTSimulator`
        The simulator wrapper, inheriting from `ast_toolbox.simulators.ASTSimulator`.
    reward_function : :py:class:`ast_toolbox.rewards.ASTReward`
        The reward function, inheriting from `ast_toolbox.rewards.ASTReward`.

    """

    def __init__(self, algo, env, n_envs=1, open_loop=True, sim=ExampleAVSimulator(), reward_function=ExampleAVReward()):

        # pdb.set_trace()
        self.open_loop = open_loop
        self.sim = sim
        self.reward_function = reward_function
        super().__init__(algo, env, n_envs)

    def obtain_samples(self, itr, batch_size=None, whole_paths=False):
        """Sample the policy for new trajectories.

        Parameters
        ----------
        itr : int
            Iteration number.
        batch_size : int
            Number of samples to be collected. If None,
            it will be default [algo.max_path_length * n_envs].
        whole_paths : bool
            Whether return all the paths or not. True
            by default. It's possible for the paths to have total actual
            sample size larger than batch_size, and will be truncated if
            this flag is true.

        Returns
        -------
        : list[dict]
            A list of sampled rollout paths.
            Each rollout path is a dictionary with the following keys:
                - observations (numpy.ndarray)
                - actions (numpy.ndarray)
                - rewards (numpy.ndarray)
                - agent_infos (dict)
                - env_infos (dict)

        """
        # pdb.set_trace()
        paths = super().obtain_samples(itr, batch_size)
        # pdb.set_trace()
        if self.open_loop:
            for path in paths:
                s_0 = path["observations"][0]

                # actions = path['env_infos']['info']['actions']
                actions = path['actions']
                # pdb.set_trace()
                end_idx, info = self.sim.simulate(actions=actions, s_0=s_0)
                # print('----- Back from simulate: ', end_idx)
                if end_idx >= 0:
                    # pdb.set_trace()
                    self.slice_dict(path, end_idx)
                rewards = self.reward_function.give_reward(
                    action=actions[end_idx],
                    info=self.sim.get_reward_info()
                )
                # print('----- Back from rewards: ', rewards)
                # pdb.set_trace()
                path["rewards"][end_idx] = rewards
                # info[:, -1] = path["rewards"][:info.shape[0]]
                # path['env_infos']['sim_info'] = info
                path['env_infos']['sim_info'] = np.zeros_like(path["rewards"])
                # import pdb; pdb.set_trace()

        return paths

    def slice_dict(self, in_dict, slice_idx):
        """Helper function to recursively parse through a dictionary of dictionaries and arrays to slice \
        the arrays at a certain index.

        Parameters
        ----------
        in_dict : dict
            Dictionary where the values are arrays or other dictionaries that follow this stipulation.
        slice_idx : int
            Index to slice each array at.

        Returns
        -------
        dict
            Dictionary where arrays at every level are sliced.

        """
        for key, value in in_dict.items():
            # pdb.set_trace()
            if isinstance(value, dict):
                in_dict[key] = self.slice_dict(value, slice_idx)
            else:
                in_dict[key][slice_idx + 1:, ...] = np.zeros_like(value[slice_idx + 1:, ...])

        return in_dict



"""Environment wrapper that runs multiple environments in parallel."""

import pickle as pickle
import uuid

from dowel import logger
import numpy as np

from garage.misc import tensor_utils
from garage.sampler.stateful_pool import singleton_pool


def worker_init_envs(g, alloc, scope, env):
    """Initialize the environment on a worker."""
    logger.log('initializing environment on worker %d' % g.worker_id)
    if not hasattr(g, 'parallel_vec_envs'):
        g.parallel_vec_envs = dict()
        g.parallel_vec_env_template = dict()
    g.parallel_vec_envs[scope] = [(idx, pickle.loads(pickle.dumps(env)))
                                  for idx in alloc]
    g.parallel_vec_env_template[scope] = env


# For these two methods below, we pack the data into batch numpy arrays
# whenever possible, to reduce communication cost


def worker_run_reset(g, flags, scope):
    """Reset the environment on a worker."""
    if not hasattr(g, 'parallel_vec_envs'):
        logger.log('on worker %d' % g.worker_id)
        import traceback
        for line in traceback.format_stack():
            logger.log(line)
        # log the stacktrace at least
        logger.log('oops')
        for k, v in g.__dict__.items():
            logger.log(str(k) + ' : ' + str(v))
        assert hasattr(g, 'parallel_vec_envs')

    assert scope in g.parallel_vec_envs
    n = len(g.parallel_vec_envs[scope])
    env_template = g.parallel_vec_env_template[scope]
    obs_dim = env_template.observation_space.flat_dim
    ret_arr = np.zeros((n, obs_dim))
    ids = []
    flat_obs = []
    reset_ids = []
    for itr_idx, (idx, env) in enumerate(g.parallel_vec_envs[scope]):
        flag = flags[idx]
        if flag:
            flat_obs.append(env.reset())
            reset_ids.append(itr_idx)
        ids.append(idx)
    if reset_ids:
        ret_arr[reset_ids] = env_template.observation_space.flatten_n(flat_obs)
    return ids, ret_arr


def worker_run_step(g, action_n, scope):
    """Step the environment on a worker."""
    assert hasattr(g, 'parallel_vec_envs')
    assert scope in g.parallel_vec_envs
    env_template = g.parallel_vec_env_template[scope]
    ids = []
    step_results = []
    for (idx, env) in g.parallel_vec_envs[scope]:
        action = action_n[idx]
        ids.append(idx)
        step_results.append(tuple(env.step(action)))
    if not step_results:
        return None
    obs, rewards, dones, env_infos = list(map(list, list(zip(*step_results))))
    obs = env_template.observation_space.flatten_n(obs)
    rewards = np.asarray(rewards)
    dones = np.asarray(dones)
    env_infos = tensor_utils.stack_tensor_dict_list(env_infos)
    return ids, obs, rewards, dones, env_infos


class ParallelVecEnvExecutor:
    """Environment wrapper that runs multiple environments in parallel."""

    def __init__(self, env, n, max_path_length, scope=None):
        if scope is None:
            # initialize random scope
            scope = str(uuid.uuid4())

        envs_per_worker = int(np.ceil(n * 1.0 / singleton_pool.n_parallel))
        alloc_env_ids = []
        rest_alloc = n
        start_id = 0
        for _ in range(singleton_pool.n_parallel):
            n_allocs = min(envs_per_worker, rest_alloc)
            alloc_env_ids.append(list(range(start_id, start_id + n_allocs)))
            start_id += n_allocs
            rest_alloc = max(0, rest_alloc - envs_per_worker)

        singleton_pool.run_each(worker_init_envs, [(alloc, scope, env)
                                                   for alloc in alloc_env_ids])

        self._alloc_env_ids = alloc_env_ids
        self._action_space = env.action_space
        self._observation_space = env.observation_space
        self._num_envs = n
        self.scope = scope
        self.ts = np.zeros(n, dtype='int')
        self.max_path_length = max_path_length

    def step(self, action_n):
        """Step all environments using the provided actions."""
        results = singleton_pool.run_each(
            worker_run_step,
            [(action_n, self.scope) for _ in self._alloc_env_ids],
        )
        results = [x for x in results if x is not None]
        ids, obs, rewards, dones, env_infos = list(zip(*results))
        ids = np.concatenate(ids)
        obs = self.observation_space.unflatten_n(np.concatenate(obs))
        rewards = np.concatenate(rewards)
        dones = np.concatenate(dones)
        env_infos = tensor_utils.split_tensor_dict_list(
            tensor_utils.concat_tensor_dict_list(env_infos))
        if env_infos is None:
            env_infos = [dict() for _ in range(self.num_envs)]

        items = list(zip(ids, obs, rewards, dones, env_infos))
        items = sorted(items, key=lambda x: x[0])

        ids, obs, rewards, dones, env_infos = list(zip(*items))

        obs = list(obs)
        rewards = np.asarray(rewards)
        dones = np.asarray(dones)

        self.ts += 1
        dones[self.ts >= self.max_path_length] = True

        reset_obs = self._run_reset(dones)
        for (i, done) in enumerate(dones):
            if done:
                obs[i] = reset_obs[i]
                self.ts[i] = 0
        return obs, rewards, dones, tensor_utils.stack_tensor_dict_list(
            list(env_infos))

    def _run_reset(self, dones):
        dones = np.asarray(dones)
        results = singleton_pool.run_each(
            worker_run_reset,
            [(dones, self.scope) for _ in self._alloc_env_ids],
        )
        ids, flat_obs = list(map(np.concatenate, list(zip(*results))))
        zipped = list(zip(ids, flat_obs))
        sorted_obs = np.asarray(
            [x[1] for x in sorted(zipped, key=lambda x: x[0])])

        done_ids, = np.where(dones)
        done_flat_obs = sorted_obs[done_ids]
        done_unflat_obs = self.observation_space.unflatten_n(done_flat_obs)
        all_obs = [None] * self.num_envs
        done_cursor = 0
        for idx, done in enumerate(dones):
            if done:
                all_obs[idx] = done_unflat_obs[done_cursor]
                done_cursor += 1
        return all_obs

    def reset(self):
        """Reset all environments."""
        dones = np.asarray([True] * self.num_envs)
        return self._run_reset(dones)

    @property
    def num_envs(self):
        """Read / write the number of environments."""
        return self._num_envs

    @property
    def action_space(self):
        """Read / write the action space."""
        return self._action_space

    @property
    def observation_space(self):
        """Read / write the observation space."""
        return self._observation_space

    def close(self):
        """Close all environments."""
        pass



import datetime
import os.path as osp
import random

import dowel
from dowel import logger as dowel_logger
import gym
import pytest
import tensorflow as tf

from garage.envs import normalize
from garage.experiment import deterministic
from garage.tf.algos import PPO
from garage.tf.baselines import GaussianCNNBaseline
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import CategoricalCNNPolicy
from tests.fixtures import snapshot_config
import tests.helpers as Rh

num_of_trials = 3

params = {
    'conv_filters': (32, 64, 64),
    'conv_filter_sizes': (5, 3, 2),
    'conv_strides': (4, 2, 1),
    'conv_pad': 'VALID',
    'hidden_sizes': (256, ),
    'n_epochs': 1000,
    'batch_size': 2048,
    'use_trust_region': True
}


class TestBenchmarkCategoricalCNNPolicy:
    '''Compare benchmarks between garage and baselines.'''

    @pytest.mark.huge
    def test_benchmark_categorical_cnn_policy(self):
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S-%f')
        benchmark_dir = './data/local/benchmarks/ppo/%s/' % timestamp
        for env_id in ['CubeCrash-v0', 'MemorizeDigits-v0']:
            env = gym.make(env_id)

            seeds = random.sample(range(100), num_of_trials)

            task_dir = osp.join(benchmark_dir, env_id)
            plt_file = osp.join(benchmark_dir,
                                '{}_benchmark_ppo.png'.format(env_id))
            relplt_file = osp.join(benchmark_dir,
                                   '{}_benchmark_ppo_mean.png'.format(env_id))
            garage_csvs = []
            garage_model_csvs = []

            for trial in range(num_of_trials):
                seed = seeds[trial]

                trial_dir = task_dir + '/trial_%d_seed_%d' % (trial + 1, seed)
                garage_dir = trial_dir + '/garage'

                with tf.Graph().as_default():
                    # Run garage algorithms
                    env.reset()
                    garage_csv = run_garage(env, seed, garage_dir)

                garage_csvs.append(garage_csv)

            env.close()

            Rh.relplot(g_csvs=garage_csvs,
                       b_csvs=garage_model_csvs,
                       g_x='Iteration',
                       g_y='AverageReturn',
                       g_z='Garage',
                       b_x='Iteration',
                       b_y='AverageReturn',
                       b_z='GarageWithModel',
                       trials=num_of_trials,
                       seeds=seeds,
                       plt_file=relplt_file,
                       env_id=env_id,
                       x_label='Iteration',
                       y_label='AverageReturn')

            Rh.plot(g_csvs=garage_csvs,
                    b_csvs=garage_model_csvs,
                    g_x='Iteration',
                    g_y='AverageReturn',
                    g_z='Garage',
                    b_x='Iteration',
                    b_y='AverageReturn',
                    b_z='GarageWithModel',
                    trials=num_of_trials,
                    seeds=seeds,
                    plt_file=plt_file,
                    env_id=env_id,
                    x_label='Iteration',
                    y_label='AverageReturn')


def run_garage(env, seed, log_dir):
    '''
    Create garage model and training.

    Replace the ppo with the algorithm you want to run.

    :param env: Environment of the task.
    :param seed: Random seed for the trial.
    :param log_dir: Log dir path.
    :return:
    '''
    deterministic.set_seed(seed)
    config = tf.ConfigProto(allow_soft_placement=True,
                            intra_op_parallelism_threads=12,
                            inter_op_parallelism_threads=12)
    sess = tf.Session(config=config)

    with LocalTFRunner(snapshot_config, sess=sess, max_cpus=12) as runner:
        env = TfEnv(normalize(env))

        policy = CategoricalCNNPolicy(
            env_spec=env.spec,
            conv_filters=params['conv_filters'],
            conv_filter_sizes=params['conv_filter_sizes'],
            conv_strides=params['conv_strides'],
            conv_pad=params['conv_pad'],
            hidden_sizes=params['hidden_sizes'])

        baseline = GaussianCNNBaseline(
            env_spec=env.spec,
            regressor_args=dict(num_filters=params['conv_filters'],
                                filter_dims=params['conv_filter_sizes'],
                                strides=params['conv_strides'],
                                padding=params['conv_pad'],
                                hidden_sizes=params['hidden_sizes'],
                                use_trust_region=params['use_trust_region']))

        algo = PPO(
            env_spec=env.spec,
            policy=policy,
            baseline=baseline,
            max_path_length=100,
            discount=0.99,
            gae_lambda=0.95,
            lr_clip_range=0.2,
            policy_ent_coeff=0.0,
            optimizer_args=dict(
                batch_size=32,
                max_epochs=10,
                tf_optimizer_args=dict(learning_rate=1e-3),
            ),
            flatten_input=False,
        )

        # Set up logger since we are not using run_experiment
        tabular_log_file = osp.join(log_dir, 'progress.csv')
        dowel_logger.add_output(dowel.StdOutput())
        dowel_logger.add_output(dowel.CsvOutput(tabular_log_file))
        dowel_logger.add_output(dowel.TensorBoardOutput(log_dir))

        runner.setup(algo, env)
        runner.train(n_epochs=params['n_epochs'],
                     batch_size=params['batch_size'])

        dowel_logger.remove_all()

        return tabular_log_file



#!/usr/bin/env python3
"""
This is an example to train a task with DDPG algorithm.

Here it creates a gym environment InvertedDoublePendulum. And uses a DDPG with
1M steps.

Results:
    AverageReturn: 250
    RiseTime: epoch 499
"""
import gym
import tensorflow as tf

from garage.experiment import run_experiment
from garage.np.exploration_strategies import OUStrategy
from garage.replay_buffer import SimpleReplayBuffer
from garage.tf.algos import DDPG
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import ContinuousMLPPolicy
from garage.tf.q_functions import ContinuousMLPQFunction


def run_task(snapshot_config, *_):
    """Run task."""
    with LocalTFRunner(snapshot_config=snapshot_config) as runner:
        env = TfEnv(gym.make('InvertedDoublePendulum-v2'))

        action_noise = OUStrategy(env.spec, sigma=0.2)

        policy = ContinuousMLPPolicy(env_spec=env.spec,
                                     hidden_sizes=[64, 64],
                                     hidden_nonlinearity=tf.nn.relu,
                                     output_nonlinearity=tf.nn.tanh)

        qf = ContinuousMLPQFunction(env_spec=env.spec,
                                    hidden_sizes=[64, 64],
                                    hidden_nonlinearity=tf.nn.relu)

        replay_buffer = SimpleReplayBuffer(env_spec=env.spec,
                                           size_in_transitions=int(1e6),
                                           time_horizon=100)

        ddpg = DDPG(env_spec=env.spec,
                    policy=policy,
                    policy_lr=1e-4,
                    qf_lr=1e-3,
                    qf=qf,
                    replay_buffer=replay_buffer,
                    target_update_tau=1e-2,
                    n_train_steps=50,
                    discount=0.9,
                    min_buffer_size=int(1e4),
                    exploration_strategy=action_noise,
                    policy_optimizer=tf.train.AdamOptimizer,
                    qf_optimizer=tf.train.AdamOptimizer)

        runner.setup(algo=ddpg, env=env)

        runner.train(n_epochs=500, n_epoch_cycles=20, batch_size=100)


run_experiment(
    run_task,
    snapshot_mode='last',
    seed=1,
)



"""Base model classes."""
import abc
from collections import namedtuple
import warnings

import tensorflow as tf


class BaseModel(abc.ABC):
    """
    Interface-only abstract class for models.

    A Model contains the structure/configuration of a set of computation
    graphs, or can be understood as a set of networks. Using a model
    requires calling `build()` with given input placeholder, which can be
    either tf.compat.v1.placeholder, or the output from another model. This
    makes composition of complex models with simple models much easier.

    Examples:
        model = SimpleModel(output_dim=2)
        # To use a model, first create a placeholder.
        # In the case of TensorFlow, we create a tf.compat.v1.placeholder.
        input_ph = tf.compat.v1.placeholder(tf.float32, shape=(None, 2))

        # Building the model
        output = model.build(input_ph)

        # We can also pass the output of a model to another model.
        # Here we pass the output from the above SimpleModel object.
        model_2 = ComplexModel(output_dim=2)
        output_2 = model_2.build(output)

    """

    def build(self, *inputs):
        """
        Output of model with the given input placeholder(s).

        This function is implemented by subclasses to create their computation
        graphs, which will be managed by Model. Generally, subclasses should
        implement `build()` directly.

        Args:
            inputs: Tensor input(s) for the model.

        Return:
            output: Tensor output(s) of the model.
        """
        pass

    @property
    def name(self):
        """Name for this Model."""
        pass

    @property
    def parameters(self):
        """
        Parameters of the Model.

        The output of a model is determined by its parameter. It could be
        the weights of a neural network model or parameters of a loss
        function model.
        """
        pass

    @parameters.setter
    def parameters(self, parameters):
        """Set parameters of the Model."""
        pass


class Network:
    """
    Network class For TensorFlow.

    A Network contains connectivity information by inputs/outputs.
    When a Network is built, it appears as a subgraph in the computation
    graphs, scoped by the Network name. All Networks built with the same
    model share the same parameters, i.e same inputs yield to same outputs.
    """

    @property
    def input(self):
        """Tensor input of the Network."""
        return self._inputs[0]

    @property
    def inputs(self):
        """Tensor inputs of the Network."""
        return self._inputs

    @property
    def output(self):
        """Tensor output of the Network."""
        return self._outputs[0]

    @property
    def outputs(self):
        """Tensor outputs of the Network."""
        return self._outputs


class Model(BaseModel):
    r"""
    Model class for TensorFlow.

    A TfModel only contains the structure/configuration of the underlying
    computation graphs. Connectivity information are all in Network class.
    A TfModel contains zero or more Network.

    When a Network is created, it reuses the parameter from the
    model and can be accessed by calling model.networks['network_name'],
    If a Network is built without given a name, the name "default" will
    be used.

    ***
    Do not call tf.global_variable_initializers() after building a model as it
    will reassign random weights to the model.
    The parameters inside a model will be initialized when calling build().
    ***

    Pickling is handled automatcailly. The target weights should be assigned to
    self._default_parameters before pickling, so that the newly created model
    can check if target weights exist or not. When unpickled, the unserialized
    model will load the weights from self._default_parameters.

    The design is illustrated as the following:

         input_1                      input_2
            |                            |
    ============== Model (TfModel)===================
    |       |                            |          |
    |       |            Parameters      |          |
    |    =============  /           \  ============ |
    |    |  default  | /             \ | Network2 | |
    |    | (Network) |/               \|(Network) | |
    |    =============                 ============ |
    |       |                            |          |
    =================================================
            |                            |
            |                            |
    (model.networks['default'].outputs)  |
                        model.networks['Network2'].outputs


    Examples are also available in tests/garage/tf/models/test_model.

    Args:
      name (str): Name of the model. It will also become the variable scope
      of the model. Every model should have a unique name.
    """

    def __init__(self, name):
        super().__init__()
        self._name = name or type(self).__name__  # name default to class
        self._networks = {}
        self._default_parameters = None
        self._variable_scope = None

    def build(self, *inputs, name=None):
        """
        Build a Network with the given input(s).

        ***
        Do not call tf.global_variable_initializers() after building a model
        as it will reassign random weights to the model.
        The parameters inside a model will be initialized when calling build().
        ***

        It uses the same, fixed variable scope for all Networks, to ensure
        parameter sharing. Different Networks must have an unique name.

        Args:
          inputs (list[tf.Tensor]) : Tensor input(s), recommended to be
            positional arguments, for example,
            def build(self, state_input, action_input, name=None).
          name (str): Name of the model, which is also the name scope of the
            model.

        Raises:
          ValueError when a Network with the same name is already built.

        Returns:
          outputs (list[tf.Tensor]): Output tensors of the model with the given
            inputs.

        """
        network_name = name or 'default'
        if not self._networks:
            # First time building the model, so self._networks are empty
            # We store the variable_scope to reenter later when we reuse it
            with tf.compat.v1.variable_scope(self._name) as vs:
                self._variable_scope = vs
                with tf.name_scope(name=network_name):
                    network = Network()
                    network._inputs = inputs
                    network._outputs = self._build(*inputs, name)
                variables = self._get_variables().values()
                tf.compat.v1.get_default_session().run(
                    tf.compat.v1.variables_initializer(variables))
                if self._default_parameters:
                    self.parameters = self._default_parameters
        else:
            if network_name in self._networks:
                raise ValueError(
                    'Network {} already exists!'.format(network_name))
            with tf.compat.v1.variable_scope(
                    self._variable_scope, reuse=True,
                    auxiliary_name_scope=False):
                with tf.name_scope(name=network_name):
                    network = Network()
                    network._inputs = inputs
                    network._outputs = self._build(*inputs, name)
        custom_in_spec = self.network_input_spec()
        custom_out_spec = self.network_output_spec()
        in_spec = ['input', 'inputs']
        out_spec = ['output', 'outputs']
        in_args = [network.input, network.inputs]
        out_args = [network.output, network.outputs]
        if isinstance(network.inputs, tuple) and len(network.inputs) > 1:
            assert len(custom_in_spec) == len(network.inputs), (
                'network_input_spec must have same length as inputs!')
            in_spec.extend(custom_in_spec)
            in_args.extend(network.inputs)
        if isinstance(network.outputs, tuple) and len(network.outputs) > 1:
            assert len(custom_out_spec) == len(network.outputs), (
                'network_output_spec must have same length as outputs!')
            out_spec.extend(custom_out_spec)
            out_args.extend(network.outputs)

        c = namedtuple(network_name, [*in_spec, *out_spec])
        all_args = in_args + out_args
        self._networks[network_name] = c(*all_args)

        return network.outputs

    def _build(self, *inputs, name=None):
        """
        Output of the model given input placeholder(s).

        User should implement _build() inside their subclassed model,
        and construct the computation graphs in this function.

        Args:
            inputs: Tensor input(s), recommended to be position arguments, e.g.
              def _build(self, state_input, action_input, name=None).
              It would be usually same as the inputs in build().
            name (str): Inner model name, also the variable scope of the
                inner model, if exist. One example is
                garage.tf.models.Sequential.

        Return:
            output: Tensor output(s) of the model.
        """
        pass

    def network_input_spec(self):
        """
        Network input spec.

        Return:
            *inputs (list[str]): List of key(str) for the network inputs.
        """
        return []

    def network_output_spec(self):
        """
        Network output spec.

        Return:
            *inputs (list[str]): List of key(str) for the network outputs.
        """
        return []

    @property
    def networks(self):
        """Networks of the model."""
        return self._networks

    @property
    def parameters(self):
        """Parameters of the model."""
        _variables = self._get_variables()
        if _variables:
            return tf.compat.v1.get_default_session().run(_variables)
        else:
            return _variables

    @parameters.setter
    def parameters(self, parameters):
        """Set model parameters."""
        variables = self._get_variables()
        for name, var in variables.items():
            if name in parameters:
                var.load(parameters[name])
            else:
                warnings.warn('No value provided for variable {}'.format(name))

    @property
    def name(self):
        """
        Name (str) of the model.

        This is also the variable scope of the model.
        """
        return self._name

    @property
    def input(self):
        """
        Default input (tf.Tensor) of the model.

        When the model is built the first time, by default it
        creates the 'default' network. This property creates
        a reference to the input of the network.
        """
        return self.networks['default'].input

    @property
    def output(self):
        """
        Default output (tf.Tensor) of the model.

        When the model is built the first time, by default it
        creates the 'default' network. This property creates
        a reference to the output of the network.
        """
        return self.networks['default'].output

    @property
    def inputs(self):
        """
        Default inputs (tf.Tensor) of the model.

        When the model is built the first time, by default it
        creates the 'default' network. This property creates
        a reference to the inputs of the network.
        """
        return self.networks['default'].inputs

    @property
    def outputs(self):
        """
        Default outputs (tf.Tensor) of the model.

        When the model is built the first time, by default it
        creates the 'default' network. This property creates
        a reference to the outputs of the network.
        """
        return self.networks['default'].outputs

    def _get_variables(self):
        if self._variable_scope:
            return {v.name: v for v in self._variable_scope.global_variables()}
        else:
            return dict()

    def __getstate__(self):
        """Object.__getstate__."""
        new_dict = self.__dict__.copy()
        del new_dict['_networks']
        new_dict['_default_parameters'] = self.parameters
        return new_dict

    def __setstate__(self, state):
        """Object.__setstate__."""
        self.__dict__.update(state)
        self._networks = {}



import collections
from copy import copy
import pickle

import dm_control.suite
import pytest

from garage.envs.dm_control import DmControlEnv
from tests.helpers import step_env


class TestDmControlEnv:
    def test_can_step(self):
        domain_name, task_name = dm_control.suite.ALL_TASKS[0]
        env = DmControlEnv.from_suite(domain_name, task_name)
        ob_space = env.observation_space
        act_space = env.action_space
        ob = env.reset()
        assert ob_space.contains(ob)
        a = act_space.sample()
        assert act_space.contains(a)
        # Skip rendering because it causes TravisCI to run out of memory
        step_env(env, render=False)
        env.close()

    @pytest.mark.nightly
    @pytest.mark.parametrize('domain_name, task_name',
                             dm_control.suite.ALL_TASKS)
    def test_all_can_step(self, domain_name, task_name):
        env = DmControlEnv.from_suite(domain_name, task_name)
        ob_space = env.observation_space
        act_space = env.action_space
        ob = env.reset()
        assert ob_space.contains(ob)
        a = act_space.sample()
        assert act_space.contains(a)
        # Skip rendering because it causes TravisCI to run out of memory
        step_env(env, render=False)
        env.close()

    def test_pickleable(self):
        domain_name, task_name = dm_control.suite.ALL_TASKS[0]
        env = DmControlEnv.from_suite(domain_name, task_name)
        round_trip = pickle.loads(pickle.dumps(env))
        assert round_trip
        # Skip rendering because it causes TravisCI to run out of memory
        step_env(round_trip, render=False)
        round_trip.close()
        env.close()

    @pytest.mark.nightly
    @pytest.mark.parametrize('domain_name, task_name',
                             dm_control.suite.ALL_TASKS)
    def test_all_pickleable(self, domain_name, task_name):
        env = DmControlEnv.from_suite(domain_name, task_name)
        round_trip = pickle.loads(pickle.dumps(env))
        assert round_trip
        # Skip rendering because it causes TravisCI to run out of memory
        step_env(round_trip, render=False)
        round_trip.close()
        env.close()

    def test_does_not_modify_actions(self):
        domain_name, task_name = dm_control.suite.ALL_TASKS[0]
        env = DmControlEnv.from_suite(domain_name, task_name)
        a = env.action_space.sample()
        a_copy = copy(a)
        env.step(a)
        if isinstance(a, collections.Iterable):
            assert a.all() == a_copy.all()
        else:
            assert a == a_copy
        env.close()

    @pytest.mark.nightly
    @pytest.mark.parametrize('domain_name, task_name',
                             dm_control.suite.ALL_TASKS)
    def test_all_does_not_modify_actions(self, domain_name, task_name):
        env = DmControlEnv.from_suite(domain_name, task_name)
        a = env.action_space.sample()
        a_copy = copy(a)
        env.step(a)
        if isinstance(a, collections.Iterable):
            assert a.all() == a_copy.all()
        else:
            assert a == a_copy
        env.close()



"""Stack frames wrapper for gym.Env."""
from collections import deque

import gym
import gym.spaces
import numpy as np


class StackFrames(gym.Wrapper):
    """gym.Env wrapper to stack multiple frames.

    Useful for training feed-forward agents on dynamic games.
    Only works with gym.spaces.Box environment with 2D single channel frames.

    Args:
        env: gym.Env to wrap.
        n_frames: number of frames to stack.

    Raises:
        ValueError: If observation space shape is not 2 or
        environment is not gym.spaces.Box.

    """

    def __init__(self, env, n_frames):
        if not isinstance(env.observation_space, gym.spaces.Box):
            raise ValueError('Stack frames only works with gym.spaces.Box '
                             'environment.')

        if len(env.observation_space.shape) != 2:
            raise ValueError(
                'Stack frames only works with 2D single channel images')

        super().__init__(env)

        self._n_frames = n_frames
        self._frames = deque(maxlen=n_frames)

        new_obs_space_shape = env.observation_space.shape + (n_frames, )
        _low = env.observation_space.low.flatten()[0]
        _high = env.observation_space.high.flatten()[0]
        self._observation_space = gym.spaces.Box(
            _low,
            _high,
            shape=new_obs_space_shape,
            dtype=env.observation_space.dtype)

    @property
    def observation_space(self):
        """gym.Env observation space."""
        return self._observation_space

    @observation_space.setter
    def observation_space(self, observation_space):
        self._observation_space = observation_space

    def _stack_frames(self):
        return np.stack(self._frames, axis=2)

    def reset(self):
        """gym.Env reset function."""
        observation = self.env.reset()
        self._frames.clear()
        for i in range(self._n_frames):
            self._frames.append(observation)

        return self._stack_frames()

    def step(self, action):
        """gym.Env step function."""
        new_observation, reward, done, info = self.env.step(action)
        self._frames.append(new_observation)

        return self._stack_frames(), reward, done, info



# Import the example classes
import os

import fire
# Useful imports
import tensorflow as tf
from garage.envs.normalized_env import normalize
from garage.experiment import run_experiment
from garage.np.baselines.linear_feature_baseline import LinearFeatureBaseline
# Import the necessary garage classes
from garage.tf.algos.ppo import PPO
from garage.tf.envs.base import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.optimizers.conjugate_gradient_optimizer import ConjugateGradientOptimizer
from garage.tf.optimizers.conjugate_gradient_optimizer import FiniteDifferenceHvp
# from garage.tf.policies.gaussian_lstm_policy import GaussianLSTMPolicy
from garage.tf.policies import GaussianLSTMPolicy

# Import the AST classes
from ast_toolbox.envs import ASTEnv
from ast_toolbox.rewards import ExampleAVReward
from ast_toolbox.samplers import ASTVectorizedSampler
from ast_toolbox.simulators import ExampleAVSimulator
from ast_toolbox.spaces import ExampleAVSpaces
from ast_toolbox.utils.go_explore_utils import load_convert_and_save_drl_expert_trajectory


def runner(
    env_args=None,
    run_experiment_args=None,
    sim_args=None,
    reward_args=None,
    spaces_args=None,
    policy_args=None,
    baseline_args=None,
    algo_args=None,
    runner_args=None,
    sampler_args=None,
    save_expert_trajectory=False,
):

    if env_args is None:
        env_args = {}

    if run_experiment_args is None:
        run_experiment_args = {}

    if sim_args is None:
        sim_args = {}

    if reward_args is None:
        reward_args = {}

    if spaces_args is None:
        spaces_args = {}

    if policy_args is None:
        policy_args = {}

    if baseline_args is None:
        baseline_args = {}

    if algo_args is None:
        algo_args = {}

    if runner_args is None:
        runner_args = {'n_epochs': 1}

    if sampler_args is None:
        sampler_args = {}

    if 'n_parallel' in run_experiment_args:
        n_parallel = run_experiment_args['n_parallel']
    else:
        n_parallel = 1
        run_experiment_args['n_parallel'] = n_parallel

    if 'max_path_length' in sim_args:
        max_path_length = sim_args['max_path_length']
    else:
        max_path_length = 50
        sim_args['max_path_length'] = max_path_length

    if 'batch_size' in runner_args:
        batch_size = runner_args['batch_size']
    else:
        batch_size = max_path_length * n_parallel
        runner_args['batch_size'] = batch_size

    def run_task(snapshot_config, *_):

        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        with tf.Session(config=config) as sess:
            with tf.variable_scope('AST', reuse=tf.AUTO_REUSE):

                with LocalTFRunner(
                        snapshot_config=snapshot_config, max_cpus=4, sess=sess) as local_runner:
                    # Instantiate the example classes
                    sim = ExampleAVSimulator(**sim_args)
                    reward_function = ExampleAVReward(**reward_args)
                    spaces = ExampleAVSpaces(**spaces_args)

                    # Create the environment
                    if 'id' in env_args:
                        env_args.pop('id')
                    env = TfEnv(normalize(ASTEnv(simulator=sim,
                                                 reward_function=reward_function,
                                                 spaces=spaces,
                                                 **env_args
                                                 )))

                    # Instantiate the garage objects
                    policy = GaussianLSTMPolicy(env_spec=env.spec, **policy_args)

                    baseline = LinearFeatureBaseline(env_spec=env.spec, **baseline_args)

                    optimizer = ConjugateGradientOptimizer
                    optimizer_args = {'hvp_approach': FiniteDifferenceHvp(base_eps=1e-5)}

                    algo = PPO(env_spec=env.spec,
                               policy=policy,
                               baseline=baseline,
                               optimizer=optimizer,
                               optimizer_args=optimizer_args,
                               **algo_args)

                    sampler_cls = ASTVectorizedSampler
                    sampler_args['sim'] = sim
                    sampler_args['reward_function'] = reward_function

                    local_runner.setup(
                        algo=algo,
                        env=env,
                        sampler_cls=sampler_cls,
                        sampler_args=sampler_args)

                    # Run the experiment
                    local_runner.train(**runner_args)

                    if save_expert_trajectory:
                        load_convert_and_save_drl_expert_trajectory(last_iter_filename=os.path.join(run_experiment_args['log_dir'],
                                                                                                    'itr_' +
                                                                                                    str(runner_args['n_epochs'] - 1) +
                                                                                                    '.pkl'),
                                                                    expert_trajectory_filename=os.path.join(run_experiment_args['log_dir'],
                                                                                                            'expert_trajectory.pkl'))

                    print('done!')

    run_experiment(
        run_task,
        **run_experiment_args,
    )


if __name__ == '__main__':
    fire.Fire()



import csv

from matplotlib import pyplot as plt

pre_path = './Data/AST/Lexington/GATRISInter/'
plot_name = "GATRISInter"
n_trial = 5

fig = plt.figure()
for i in range(n_trial):
    csv_path = pre_path + str(i) + '/process.csv'
    plt.subplot(2, 3, i + 1)
    with open(csv_path) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        SumOtherWeights = []
        iterations = []
        for (i, row) in enumerate(csv_reader):
            print(i)
            if i == 0:
                entry_dict = {}
                for index in range(len(row)):
                    entry_dict[row[index]] = index
            else:
                SumOtherWeights.append(float(row[entry_dict["SumOtherWeights"]]))
                iterations.append(int(row[entry_dict["Iteration"]]))

    plt.scatter(iterations, SumOtherWeights, s=0.1, marker=',')
    plt.xlabel('Iteration')
    plt.ylabel('SumOtherWeights')
fig.savefig('Data/Plot/sumOtherWeights/' + plot_name + '_SumOtherWeights.pdf')
plt.close(fig)



import numpy as np
import tensorflow as tf

from garage.tf.distributions.base import Distribution
from garage.tf.distributions.categorical import Categorical

TINY = 1e-8


class RecurrentCategorical(Distribution):
    def __init__(self, dim, name='RecurrentCategorical'):
        self._cat = Categorical(dim, name)
        self._dim = dim
        self._name = name

    @property
    def dim(self):
        return self._dim

    def kl_sym(self, old_dist_info_vars, new_dist_info_vars, name=None):
        """
        Compute the symbolic KL divergence of two categorical distributions
        """
        with tf.name_scope(name, 'kl_sym',
                           [old_dist_info_vars, new_dist_info_vars]):
            old_prob_var = old_dist_info_vars['prob']
            new_prob_var = new_dist_info_vars['prob']
            # Assume layout is N * T * A
            return tf.reduce_sum(
                old_prob_var * (tf.math.log(old_prob_var + TINY) -
                                tf.math.log(new_prob_var + TINY)),
                axis=2)

    def kl(self, old_dist_info, new_dist_info):
        """
        Compute the KL divergence of two categorical distributions
        """
        old_prob = old_dist_info['prob']
        new_prob = new_dist_info['prob']
        return np.sum(
            old_prob * (np.log(old_prob + TINY) - np.log(new_prob + TINY)),
            axis=2)

    def likelihood_ratio_sym(self,
                             x_var,
                             old_dist_info_vars,
                             new_dist_info_vars,
                             name=None):
        with tf.name_scope(name, 'likelihood_ratio_sym',
                           [x_var, old_dist_info_vars, new_dist_info_vars]):
            old_prob_var = old_dist_info_vars['prob']
            new_prob_var = new_dist_info_vars['prob']
            # Assume layout is N * T * A
            a_dim = tf.shape(x_var)[2]
            flat_ratios = self._cat.likelihood_ratio_sym(
                tf.reshape(x_var, tf.stack([-1, a_dim])),
                dict(prob=tf.reshape(old_prob_var, tf.stack([-1, a_dim]))),
                dict(prob=tf.reshape(new_prob_var, tf.stack([-1, a_dim]))))
            return tf.reshape(flat_ratios, tf.shape(old_prob_var)[:2])

    def entropy(self, dist_info):
        probs = dist_info['prob']
        return -np.sum(probs * np.log(probs + TINY), axis=2)

    def entropy_sym(self, dist_info_vars, name=None):
        with tf.name_scope(name, 'entropy_sym', [dist_info_vars]):
            probs = dist_info_vars['prob']
            return -tf.reduce_sum(probs * tf.math.log(probs + TINY), 2)

    def log_likelihood_sym(self, xs, dist_info_vars, name=None):
        with tf.name_scope(name, 'log_likelihood_sym', [xs, dist_info_vars]):
            probs = dist_info_vars['prob']
            # Assume layout is N * T * A
            a_dim = tf.shape(probs)[2]
            flat_logli = self._cat.log_likelihood_sym(
                tf.reshape(xs, tf.stack([-1, a_dim])),
                dict(prob=tf.reshape(probs, tf.stack((-1, a_dim)))))
            return tf.reshape(flat_logli, tf.shape(probs)[:2])

    def log_likelihood(self, xs, dist_info):
        probs = dist_info['prob']
        # Assume layout is N * T * A
        a_dim = tf.shape(probs)[2]
        flat_logli = self._cat.log_likelihood_sym(
            xs.reshape((-1, a_dim)), dict(prob=probs.reshape((-1, a_dim))))
        return flat_logli.reshape(probs.shape[:2])

    @property
    def dist_info_specs(self):
        return [('prob', (self.dim, ))]



"""Module for parallel sampling a batch of rollouts"""
import numpy as np
import tensorflow as tf
from garage.sampler.base import BaseSampler
from garage.sampler.stateful_pool import singleton_pool
from garage.sampler.utils import truncate_paths

from ast_toolbox.rewards import ExampleAVReward
from ast_toolbox.samplers import parallel_sampler
from ast_toolbox.simulators import ExampleAVSimulator


def worker_init_tf(g):
    """Initialize the tf.Session on a worker.

    Parameters
    ----------
    g : :py:class:`garage.sampler.stateful_pool.SharedGlobal`
        SharedGlobal class from :py:mod:`garage.sampler.stateful_pool`.
    """
    g.sess = tf.compat.v1.Session()
    g.sess.__enter__()


def worker_init_tf_vars(g):
    """Initialize the policy parameters on a worker.

    Parameters
    ----------
    g : :py:class:`garage.sampler.stateful_pool.SharedGlobal`
        SharedGlobal class from :py:mod:`garage.sampler.stateful_pool`.
    """
    g.sess.run(tf.compat.v1.global_variables_initializer())


class BatchSampler(BaseSampler):
    """Collects samples in parallel using a stateful pool of workers.

    Parameters
    ----------
    algo : :py:class:`garage.np.algos.base.RLAlgorithm`
        The algorithm.
    env : :py:class:`ast_toolbox.envs.ASTEnv`
        The environment.
    n_envs : int
        Number of parallel environments to run.
    open_loop : bool
        True if the simulation is open-loop, meaning that AST must generate all actions ahead of time, instead
        of being able to output an action in sync with the simulator, getting an observation back before
        the next action is generated. False to get interactive control, which requires that `blackbox_sim_state`
        is also False.
    batch_simulate : bool
        When in `obtain_samples` with `open_loop == True`, the sampler will call `self.sim.batch_simulate_paths` if
        `batch_simulate` is True, and `self.sim.simulate` if False.
    sim : :py:class:`ast_toolbox.simulators.ASTSimulator`
        The simulator wrapper, inheriting from `ast_toolbox.simulators.ASTSimulator`.
    reward_function : :py:class:`ast_toolbox.rewards.ASTReward`
        The reward function, inheriting from `ast_toolbox.rewards.ASTReward`.
    Args:
        algo (garage.np.algos.RLAlgorithm): The algorithm.
        env (gym.Env): The environment.

    """

    def __init__(self, algo, env, n_envs=1, open_loop=True, batch_simulate=False,
                 sim=ExampleAVSimulator(), reward_function=ExampleAVReward()):
        """


        """
        # pdb.set_trace()
        super(BatchSampler, self).__init__(algo, env)
        self.n_envs = n_envs
        self.open_loop = open_loop
        self.sim = sim
        self.reward_function = reward_function
        self.open_loop = open_loop
        self.batch_simulate = batch_simulate

    def start_worker(self):
        """Initialize the sampler."""
        assert singleton_pool.initialized, (
            'Use singleton_pool.initialize(n_parallel) to setup workers.')
        if singleton_pool.n_parallel > 1:
            singleton_pool.run_each(worker_init_tf)
        parallel_sampler.populate_task(self.env, self.algo.policy)
        if singleton_pool.n_parallel > 1:
            singleton_pool.run_each(worker_init_tf_vars)

    # def shutdown_worker(self):
    #     """Shutdown worker function."""
    #     parallel_sampler.terminate_task(scope=self.algo.scope)

    def shutdown_worker(self):
        """Terminate workers if necessary."""
        parallel_sampler.terminate_task(scope=self.algo.scope)

    def obtain_samples(self, itr, batch_size=None, whole_paths=True):
        """Collect samples for the given iteration number.

        Parameters
        ----------
        itr : int
            Iteration number.
        batch_size : int, optional
            How many simulation steps to run in each epoch.
        whole_paths : bool, optional
            Whether to return the full rollout paths data.
        """
        if not batch_size:
            batch_size = self.algo.max_path_length * self.n_envs

        # cur_params = self.algo.policy.get_param_values()
        cur_policy_params = self.algo.policy.get_param_values()
        cur_env_params = self.algo.env.get_param_values()
        paths = parallel_sampler.sample_paths(
            policy_params=cur_policy_params,
            max_samples=batch_size,
            max_path_length=self.algo.max_path_length,
            env_params=cur_env_params,
            scope=self.algo.scope,
        )
        # TODO: Doing the path correction here means the simulations will not be parallel.
        #  Need to make own parallel sampler and put it there to make that work
        if self.open_loop:
            if self.batch_simulate:
                # import pdb; pdb.set_trace()
                paths = self.sim.batch_simulate_paths(paths=paths, reward_function=self.reward_function)
            else:
                for path in paths:
                    s_0 = path["observations"][0]

                    # actions = path['env_infos']['info']['actions']
                    actions = path['actions']
                    # pdb.set_trace()
                    end_idx, info = self.sim.simulate(actions=actions, s_0=s_0)
                    # print('----- Back from simulate: ', end_idx)
                    if end_idx >= 0:
                        # pdb.set_trace()
                        self.slice_dict(path, end_idx)
                    rewards = self.reward_function.give_reward(
                        action=actions[end_idx],
                        info=self.sim.get_reward_info()
                    )
                    # print('----- Back from rewards: ', rewards)
                    # pdb.set_trace()
                    path["rewards"][end_idx] = rewards
                    # info[:, -1] = path["rewards"][:info.shape[0]]
                    # path['env_infos']['cache'] = info
                    path['env_infos']['cache'] = np.zeros_like(path["rewards"])
                    # import pdb; pdb.set_trace()

        # return paths if whole_paths else truncate_paths(paths, batch_size)
        if whole_paths:
            return paths
        else:
            paths_truncated = truncate_paths(paths, batch_size)
            return paths_truncated

    def slice_dict(self, in_dict, slice_idx):
        """Helper function to recursively parse through a dictionary of dictionaries and arrays to slice \
        the arrays at a certain index.

        Parameters
        ----------
        in_dict : dict
            Dictionary where the values are arrays or other dictionaries that follow this stipulation.
        slice_idx : int
            Index to slice each array at.

        Returns
        -------
        dict
            Dictionary where arrays at every level are sliced.

        """
        for key, value in in_dict.items():
            # pdb.set_trace()
            if isinstance(value, dict):
                in_dict[key] = self.slice_dict(value, slice_idx)
            else:
                in_dict[key][slice_idx + 1:, ...] = np.zeros_like(value[slice_idx + 1:, ...])

        return in_dict
